@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2004-2012 Andrew Mihal
+ * Copyright (C) 2004-2009 Andrew Mihal
+ * Copyright (C) 2009-2016 Christoph Spiel
  *
  * This file is part of Enblend.
  *
@@ -34,12 +35,12 @@
 #include <vigra/copyimage.hxx>
 #include <vigra/imageinfo.hxx>
 #include <vigra/impex.hxx>
+#include <vigra/impexalpha.hxx>
 #include <vigra/inspectimage.hxx>
 #include <vigra/numerictraits.hxx>
 #include <vigra/transformimage.hxx>
 
-#include "vigra_ext/functoraccessor.hxx"
-#include "vigra_ext/impexalpha.hxx"
+#include "functoraccessor.hxx"
 
 #include "common.h"
 #include "fixmath.h"
@@ -66,13 +67,13 @@ exportImagePreferablyWithAlpha(const ImageType* image,
                                const vigra::ImageExportInfo& outputImageInfo)
 {
     try {
-        vigra_ext::exportImageAlpha(srcImageRange(*image),
-                                    srcIter(mask->upperLeft(), mask_accessor),
-                                    outputImageInfo);
+        vigra::exportImageAlpha(srcImageRange(*image),
+                                srcIter(mask->upperLeft(), mask_accessor),
+                                outputImageInfo);
     } catch (std::exception& e) {
+        std::cerr << command << ": warning: must fall back to export image without alpha channel" << std::endl;
 #ifdef DEBUG
-        std::cerr << "+ exportImagePreferablyWithAlpha: fallback to export without alpha channel\n";
-        std::cerr << "+ exportImagePreferablyWithAlpha: because of " << e.what() << "\n";
+        std::cerr << "+ exportImagePreferablyWithAlpha: exception description follows..." << e.what();
 #endif
         exportImage(srcImageRange(*image), outputImageInfo);
     }
@@ -115,7 +116,7 @@ checkpoint(const std::pair<ImageType*, AlphaType*>& p,
         vigra::NumericTraits<ImagePixelComponentType>::max() :
         1.0;
 #ifdef DEBUG
-    std::cerr << "+ checkpoint: input range:  ("
+    std::cerr << "+ checkpoint: input range: ("
               << static_cast<double>(inputMin) << ", "
               << static_cast<double>(inputMax) << ")\n"
               << "+ checkpoint: output range: ("
@@ -133,24 +134,22 @@ checkpoint(const std::pair<ImageType*, AlphaType*>& p,
 #endif
             exportImagePreferablyWithAlpha(image, mask, ata, outputImageInfo);
         } else {
-            std::string pixel_type(outputImageInfo.getPixelType());
-            boost::algorithm::to_lower(pixel_type);
+            const std::string pixel_type(enblend::to_lower_copy(std::string(outputImageInfo.getPixelType())));
             std::cerr << command
                       << ": info: narrowing channel width for output as \""
                       << pixel_type << "\"" << std::endl;
 
             ImageType lowDepthImage(image->width(), image->height());
-            transformImageMP(srcImageRange(*image),
-                             destImage(lowDepthImage),
-                             vigra::linearRangeMapping(ImagePixelType(inputMin),
-                                                       ImagePixelType(inputMax),
-                                                       ImagePixelType(outputRange.first),
-                                                       ImagePixelType(outputRange.second)));
+            vigra::omp::transformImage(srcImageRange(*image),
+                                       destImage(lowDepthImage),
+                                       vigra::linearRangeMapping(ImagePixelType(inputMin),
+                                                                 ImagePixelType(inputMax),
+                                                                 ImagePixelType(outputRange.first),
+                                                                 ImagePixelType(outputRange.second)));
             exportImagePreferablyWithAlpha(&lowDepthImage, mask, ata, outputImageInfo);
         }
     } else {
-        std::string pixel_type(outputImageInfo.getPixelType());
-        boost::algorithm::to_lower(pixel_type);
+        const std::string pixel_type(enblend::to_lower_copy(std::string(outputImageInfo.getPixelType())));
         std::cerr << command
                   << ": info: rescaling floating-point data for output as \""
                   << pixel_type << "\"" << std::endl;
@@ -169,12 +168,12 @@ checkpoint(const std::pair<ImageType*, AlphaType*>& p,
                  AlphaTraits<IntegralPixelComponentType>::zero()),
                 mask->accessor());
 
-        transformImageMP(srcImageRange(*image),
-                         destImage(integralImage),
-                         vigra::linearRangeMapping(ImagePixelType(inputMin),
-                                                   ImagePixelType(inputMax),
-                                                   IntegralPixelType(outputRange.first),
-                                                   IntegralPixelType(outputRange.second)));
+        vigra::omp::transformImage(srcImageRange(*image),
+                                   destImage(integralImage),
+                                   vigra::linearRangeMapping(ImagePixelType(inputMin),
+                                                             ImagePixelType(inputMax),
+                                                             IntegralPixelType(outputRange.first),
+                                                             IntegralPixelType(outputRange.second)));
         exportImagePreferablyWithAlpha(&integralImage, mask, ata, outputImageInfo);
     }
 }
@@ -207,7 +206,7 @@ import(const vigra::ImageImportInfo& info,
                  AlphaTraits<AlphaPixelType>::max()),
                 alpha.second);
 
-        vigra_ext::importImageAlpha(info, image, vigra::destIter(alpha.first, ata));
+        vigra::importImageAlpha(info, image, vigra::destIter(alpha.first, ata));
     } else {
         // Import image without alpha.  Initialize the alpha image to 100%.
         importImage(info, image.first, image.second);
@@ -226,12 +225,12 @@ import(const vigra::ImageImportInfo& info,
         static_cast<double>(vigra::NumericTraits<ImagePixelComponentType>::max()) :
         1.0;
     if (inputRange.first != min || inputRange.second != max) {
-        transformImageMP(srcIterRange(image.first, image.first + extent, image.second),
-                         vigra::destIter(image.first, image.second),
-                         vigra::linearRangeMapping(ImagePixelType(inputRange.first),
-                                                   ImagePixelType(inputRange.second),
-                                                   ImagePixelType(min),
-                                                   ImagePixelType(max)));
+        vigra::omp::transformImage(srcIterRange(image.first, image.first + extent, image.second),
+                                   vigra::destIter(image.first, image.second),
+                                   vigra::linearRangeMapping(ImagePixelType(inputRange.first),
+                                                             ImagePixelType(inputRange.second),
+                                                             ImagePixelType(min),
+                                                             ImagePixelType(max)));
     }
 }
 
@@ -251,8 +250,8 @@ assemble(std::list<vigra::ImageImportInfo*>& imageInfoList, vigra::Rect2D& input
 
     // No more images to assemble?
     if (imageInfoList.empty()) {
-        return std::pair<ImageType*, AlphaType*>(static_cast<ImageType*>(NULL),
-                                                 static_cast<AlphaType*>(NULL));
+        return std::pair<ImageType*, AlphaType*>(static_cast<ImageType*>(nullptr),
+                                                 static_cast<AlphaType*>(nullptr));
     }
 
     // Create an image to assemble input images into.
@@ -292,19 +291,19 @@ assemble(std::list<vigra::ImageImportInfo*>& imageInfoList, vigra::Rect2D& input
             vigra::ImageImportInfo* info = *i;
 
             // Load the next image.
-            ImageType* src = new ImageType(info->size());
-            AlphaType* srcA = new AlphaType(info->size());
+            std::unique_ptr<ImageType> src {new ImageType(info->size())};
+            std::unique_ptr<AlphaType> srcA {new AlphaType(info->size())};
 
             import(*info, destImage(*src), destImage(*srcA));
 
             // Check for overlap.
             bool overlapFound = false;
-            AlphaIteratorType dy =
-                imageA->upperLeft() - inputUnion.upperLeft() + info->getPosition();
+            AlphaIteratorType dy = imageA->upperLeft() - inputUnion.upperLeft() + info->getPosition();
             AlphaAccessor da = imageA->accessor();
             AlphaIteratorType sy = srcA->upperLeft();
             AlphaIteratorType send = srcA->lowerRight();
             AlphaAccessor sa = srcA->accessor();
+
             for(; sy.y < send.y; ++sy.y, ++dy.y) {
                 AlphaIteratorType sx = sy;
                 AlphaIteratorType dx = dy;
@@ -328,19 +327,28 @@ assemble(std::list<vigra::ImageImportInfo*>& imageInfoList, vigra::Rect2D& input
                 }
 
                 const vigra::Diff2D srcPos = info->getPosition();
-                vigra::copyImageIf(srcImageRange(*src),
-                                   maskImage(*srcA),
-                                   vigra::destIter(image->upperLeft() - inputUnion.upperLeft() + srcPos));
-                vigra::copyImageIf(srcImageRange(*srcA),
-                                   maskImage(*srcA),
-                                   vigra::destIter(imageA->upperLeft() - inputUnion.upperLeft() + srcPos));
+#ifdef OPENMP
+                omp::scoped_nested(true);
+                omp::scoped_dynamic(true);
+#pragma omp parallel
+                {
+#pragma omp single nowait
+                    {
+#endif
+                        vigra::omp::copyImageIf(srcImageRange(*src),
+                                                maskImage(*srcA),
+                                                vigra::destIter(image->upperLeft() - inputUnion.upperLeft() + srcPos));
+                        vigra::omp::copyImageIf(srcImageRange(*srcA),
+                                                maskImage(*srcA),
+                                                vigra::destIter(imageA->upperLeft() - inputUnion.upperLeft() + srcPos));
+#ifdef OPENMP
+                    } // omp single
+                } // omp parallel
+#endif
 
                 // Remove info from list later.
                 toBeRemoved.push_back(i);
             }
-
-            delete src;
-            delete srcA;
         }
 
         // Erase the ImageImportInfos we used.
