@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2004-2012 Andrew Mihal
+ * Copyright (C) 2004-2009 Andrew Mihal
+ * Copyright (C) 2009-2016 Christoph Spiel
  *
  * This file is part of Enblend.
  *
@@ -89,21 +90,33 @@ blend(std::vector<MaskPyramidType*>* maskGP,
         std::cerr.flush();
     }
 
-    for (unsigned int layer = 0; layer < maskGP->size(); layer++) {
-        if (Verbose >= VERBOSE_BLEND_MESSAGES) {
-            std::cerr << " l" << layer;
-            std::cerr.flush();
-        }
+#ifdef OPENMP
+    omp::scoped_nested(true);
+    omp::scoped_dynamic(true);
+#pragma omp parallel
+    {
+#pragma omp single nowait
+        {
+#endif
+            for (unsigned int layer = 0; layer < maskGP->size(); layer++) {
+                if (Verbose >= VERBOSE_BLEND_MESSAGES) {
+                    std::cerr << " l" << layer;
+                    std::cerr.flush();
+                }
 
-        combineThreeImagesMP(srcImageRange(*((*maskGP)[layer])),
-                             srcImage(*((*whiteLP)[layer])),
-                             srcImage(*((*blackLP)[layer])),
-                             destImage(*((*blackLP)[layer])),
-                             CartesianBlendFunctor<typename MaskPyramidType::value_type>(maskPyramidWhiteValue));
-    }
+                vigra::omp::combineThreeImages(srcImageRange(*((*maskGP)[layer])),
+                                               srcImage(*((*whiteLP)[layer])),
+                                               srcImage(*((*blackLP)[layer])),
+                                               destImage(*((*blackLP)[layer])),
+                                               CartesianBlendFunctor<typename MaskPyramidType::value_type>(maskPyramidWhiteValue));
+            }
+#ifdef OPENMP
+        } // omp single
+    } // omp parallel
+#endif
 
     if (Verbose >= VERBOSE_BLEND_MESSAGES) {
-        std::cerr << endl;
+        std::cerr << std::endl;
     }
 }
 
