@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2004-2012 Andrew Mihal
+ * Copyright (C) 2004-2009 Andrew Mihal
+ * Copyright (C) 2009-2016 Christoph Spiel
  *
  * This file is part of Enblend.
  *
@@ -25,6 +26,7 @@
 #endif
 
 #include "common.h"
+#include "parameter.h"
 #include "pyramid.h"
 
 
@@ -117,7 +119,8 @@ roiBounds(const vigra::Rect2D& inputUnion,
 
     // Verify the number of levels based on the size of the ROI.
     unsigned int roiShortDimension = std::min(roiBB.width(), roiBB.height());
-    const unsigned int minimumPyramidLevels = 1U; //< src::minimum-pyramid-levels 1
+    const unsigned int minimumPyramidLevels =
+        parameter::as_unsigned("minimum-pyramid-levels", 1U); //< minimum-pyramid-levels 1
     unsigned int allowableLevels = minimumPyramidLevels;
     while (allowableLevels <= MAX_PYRAMID_LEVELS) {
         if (roiShortDimension <= 8U) {
@@ -135,7 +138,7 @@ roiBounds(const vigra::Rect2D& inputUnion,
         if (ExactLevels >= 1) {
             if (ExactLevels > static_cast<int>(allowableLevels)) {
                 std::cerr << command << ": warning: cannot blend with " << ExactLevels << " pyramid level(s) as\n"
-                          << command << ": warning:     image geometry precludes using more than "
+                          << command << ": warning: image geometry precludes using more than "
                           << allowableLevels << " pyramid level(s)" << std::endl;
             }
             allowableLevels = std::min(allowableLevels, static_cast<unsigned int>(ExactLevels));
@@ -145,7 +148,7 @@ roiBounds(const vigra::Rect2D& inputUnion,
             } else {
                 std::cerr << ": warning: cannot sensibly blend with " << allowableLevels << ExactLevels
                           << " levels\n"
-                          << command << ": warning:     will not use less than " << minimumPyramidLevels
+                          << command << ": warning: will not use less than " << minimumPyramidLevels
                           << " pyramid level(s)" << std::endl;
                 allowableLevels = minimumPyramidLevels;
             }
@@ -153,7 +156,13 @@ roiBounds(const vigra::Rect2D& inputUnion,
     }
 
     if (Verbose >= VERBOSE_PYRAMID_MESSAGES) {
-        std::cerr << command << ": info: using " << allowableLevels << " blending level(s)" << std::endl;
+        std::cerr << command << ": info: using ";
+        if (allowableLevels == 1) {
+            std::cerr << "single blending level";
+        } else {
+            std::cerr << allowableLevels << " blending levels";
+        }
+        std::cerr << std::endl;
     }
 
     assert(allowableLevels >= minimumPyramidLevels);
