@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2004-2012 Andrew Mihal
+ * Copyright (C) 2004-2009 Andrew Mihal
+ * Copyright (C) 2009-2016 Christoph Spiel
  *
  * This file is part of Enblend.
  *
@@ -25,35 +26,31 @@
 #endif
 
 #include <algorithm>
+#include <cassert>
 #include <fstream>
 #include <iomanip>
 #include <limits>
+#include <locale>
 #include <map>
 #include <stdexcept>
 #include <string>
 #include <vector>
 
-#include <boost/algorithm/string/trim.hpp>
-#include <boost/assign/list_inserter.hpp>
-#include <boost/assign/list_of.hpp>
-#include <boost/lambda/lambda.hpp>
-#include <boost/scoped_ptr.hpp>
-
 #include <vigra/numerictraits.hxx>
 
 #include "error_message.h"
 #include "filenameparse.h"
 
-#define NUMERIC_OPTION_DELIMITERS ";:/"
-#define PATH_OPTION_DELIMITERS ",;:"
-#define ASSIGNMENT_CHARACTERS "="
+#define NUMERIC_OPTION_DELIMITERS ";:/"    //< numeric-option-delimiters ;:/
+#define PATH_OPTION_DELIMITERS ",;:"       //< path-option-delimiters ,;:
+#define ASSIGNMENT_CHARACTERS "="          //< assignment-characters =
 
 #define MASK_COMPRESSION "DEFLATE"
 
 // IMPLEMENTATION NOTE: For 30 or more pyramid levels, the full width
 // will just barely fit in a 32-bit integer.  When this range is added
 // to a bounding box it will certainly overflow the vigra::Diff2D.
-#define MAX_PYRAMID_LEVELS 29   //< src::maximum-pyramid-levels 29
+#define MAX_PYRAMID_LEVELS 29   //< maximum-pyramid-levels 29
 
 // Colors used in the optimizer visualization
 #define VISUALIZE_RGB_COLOR_BLUE1    vigra::RGBValue<vigra::UInt8>(  0,   0, 255)
@@ -113,36 +110,36 @@ typedef enum
 } marker_t;
 
 
-//< src::visualize-movable-point light orange
+//< visualize-movable-point light orange
 #define VISUALIZE_MOVABLE_POINT VISUALIZE_RGB_COLOR_ORANGE1
-//< src::mark-movable-point diamond
+//< mark-movable-point diamond
 #define MARK_MOVABLE_POINT HOLLOW_DIAMOND_MARKER
-//< src::visualize-frozen-point bright white
+//< visualize-frozen-point bright white
 #define VISUALIZE_FROZEN_POINT VISUALIZE_RGB_COLOR_GRAY255
-//< src::mark-frozen-point cross
+//< mark-frozen-point cross
 #define MARK_FROZEN_POINT CROSS_MARKER
-//< src::visualize-initial-path-color dark yellow
+//< visualize-initial-path-color dark yellow
 #define VISUALIZE_INITIAL_PATH       VISUALIZE_RGB_COLOR_YELLOW4
-//< src::visualize-short-path-value-color bright yellow
+//< visualize-short-path-value-color bright yellow
 #define VISUALIZE_SHORT_PATH_VALUE   VISUALIZE_RGB_COLOR_YELLOW1
-//< src::visualize-first-vertex-value-color medium green
+//< visualize-first-vertex-value-color medium green
 #define VISUALIZE_FIRST_VERTEX_VALUE VISUALIZE_RGB_COLOR_GREEN3
-//< src::visualize-next-vertex-value-color light green
+//< visualize-next-vertex-value-color light green
 #define VISUALIZE_NEXT_VERTEX_VALUE  VISUALIZE_RGB_COLOR_GREEN2
-//< src::visualize-no-overlap-value-color dark red
+//< visualize-no-overlap-value-color dark red
 #define VISUALIZE_NO_OVERLAP_VALUE   VISUALIZE_RGB_COLOR_RED4
-//< src::visualize-state-space-color dark blue
+//< visualize-state-space-color dark blue
 #define VISUALIZE_STATE_SPACE        VISUALIZE_RGB_COLOR_BLUE3
-//< src::visualize-state-space-inside-color bright cyan
+//< visualize-state-space-inside-color bright cyan
 #define VISUALIZE_STATE_SPACE_INSIDE VISUALIZE_RGB_COLOR_CYAN1
-//< src::visualize-state-space-unconverged-color bright magenta
+//< visualize-state-space-unconverged-color bright magenta
 #define VISUALIZE_STATE_SPACE_UNCONVERGED VISUALIZE_RGB_COLOR_MAGENTA1
 
 
-// For CIECAM blending, Enblend and Enfuse transform the input images
-// from their respective RGB color spaces to XYZ space (and then on to
-// JCh).  The following two #defines control the color transformation
-// from and to XYZ space.
+// For CIELAB, CIELUV, and CIECAM blending, Enblend and Enfuse
+// transform the input images from their respective RGB color spaces
+// to XYZ space (and for CIECAM then on to JCh).  The following two
+// #defines control the color transformation from and to XYZ space.
 #define RENDERING_INTENT_FOR_BLENDING INTENT_PERCEPTUAL
 #define TRANSFORMATION_FLAGS_FOR_BLENDING cmsFLAGS_NOCACHE
 
@@ -150,6 +147,7 @@ typedef enum
 // Select our preferred type of image depending on what ./configure
 // tells us.
 #ifdef CACHE_IMAGES
+#error "The ImageCache feature has been withdrawn.  Reconfigure without ImageCache and re-build."
 #define IMAGETYPE vigra_ext::CachedFileImage
 #else
 #define IMAGETYPE vigra::BasicImage
@@ -160,12 +158,16 @@ typedef enum
 #define sleep(m_duration) Sleep(m_duration)
 #endif
 
+#if defined __MINGW32__ || defined __MINGW64__
+#undef rand_r
+#undef strtok_r
+#endif
 
 #define PENALIZE_DEPRECATED_OPTION(m_old_name, m_new_name) \
     do { \
-        cerr << command << \
+        std::cerr << command <<                                         \
             ": info: option \"" m_old_name "\" is deprecated; use \"" m_new_name "\" instead" << \
-            endl; \
+            std::endl;                                                  \
         sleep(1); \
     } while (false)
 
@@ -177,10 +179,18 @@ typedef enum
 class never_reached : public std::runtime_error
 {
 public:
-    never_reached(const std::string& a_message) : std::runtime_error(a_message) {}
-    virtual ~never_reached() throw() {}
+    never_reached() = delete;
+    explicit never_reached(const std::string& a_message) : std::runtime_error(a_message) {}
+    virtual ~never_reached() noexcept {}
 };
 
+#ifdef __GNUC__
+#define NEVER_REACHED(m_message) \
+    do {__builtin_unreachable();  throw ::never_reached(m_message);} while (false)
+#else
+#define NEVER_REACHED(m_message) throw ::never_reached(m_message)
+#endif
+
 
 namespace enblend {
 
@@ -216,58 +226,11 @@ square(t x)
 }
 
 
-/** Answer the previous iterator of i. */
-template <class iterator>
-inline iterator
-prev(iterator i)
-{
-    return --i;
-}
-
-
-/** Answer the following iterator of i. */
-template <class iterator>
-inline iterator
-next(iterator i)
-{
-    return ++i;
-}
-
-
 /** Test whether s starts with p. */
-inline bool
+inline static bool
 starts_with(const std::string& s, const std::string& p)
 {
-    return s.substr(0, p.length()) == p;
-}
-
-
-/** String tokenizer similar to strtok_r().
- *  In contrast to strtok_r this function returns an empty string for
- *  each pair of successive delimiters.  Function strtok_r skips them.
- */
-char*
-strtoken_r(char* str, const char* delim, char** save_ptr)
-{
-    char *s = str ? str : *save_ptr;
-
-    if (s)
-    {
-        char *token = s;
-
-        while (*s != 0 && !strchr(delim, (int) *s))
-        {
-            s++;
-        }
-        *save_ptr = *s == 0 ? NULL : s + 1;
-        *s = 0;
-
-        return token;
-    }
-    else
-    {
-        return NULL;
-    }
+    return s.substr(0U, p.length()) == p;
 }
 
 
@@ -312,9 +275,7 @@ can_open_file(const std::string& aFilename)
 std::string
 getFileType(const std::string& aFileName)
 {
-    std::string ext = aFileName.substr(aFileName.rfind(".") + 1);
-
-    boost::algorithm::to_upper(ext);
+    const std::string ext(to_upper_copy(aFileName.substr(aFileName.rfind(".") + 1U)));
 
     if (ext == "JPG") return "JPEG";
     else if (ext == "TIF") return "TIFF";
@@ -329,9 +290,7 @@ getFileType(const std::string& aFileName)
 boundary_t
 wraparoundOfString(const char* aWraparoundMode)
 {
-    std::string mode = std::string(aWraparoundMode);
-
-    boost::algorithm::to_upper(mode);
+    const std::string mode(to_upper_copy(std::string(aWraparoundMode)));
 
     if (mode == "NONE" || mode == "OPEN") return OpenBoundaries;
     else if (mode == "HORIZONTAL") return HorizontalStrip;
@@ -358,7 +317,7 @@ stringOfWraparound(boundary_t aBoundaryMode)
     case DoubleStrip:
         return "both";
     default:
-        throw never_reached("switch control expression \"aBoundaryMode\" out of range");
+        NEVER_REACHED("switch control expression \"aBoundaryMode\" out of range");
     }
 }
 
@@ -430,8 +389,8 @@ numberOfString(const char* a_string,                // string we want to convert
             if (long_int_value < traits::min() || long_int_value > traits::max())
             {
                 std::cerr << command << ": "
-                          << "signed number " << long_int_value
-                          << " out of range [" << traits::min() << " .. " << traits::max() << "]"
+                          << "signed number x = " << long_int_value
+                          << " out of range " << traits::min() << " <= x <= " << traits::max()
                           << std::endl;
                 exit(1);
             }
@@ -445,8 +404,8 @@ numberOfString(const char* a_string,                // string we want to convert
             if (long_int_value < 0L || long_int_value > traits::max())
             {
                 std::cerr << command << ": "
-                          << "unsigned number " << long_int_value
-                          << " out of range [0 .. " << traits::max() << "]"
+                          << "unsigned number x = " << long_int_value
+                          << " out of range 0 <= x <= " << traits::max()
                           << std::endl;
                 exit(1);
             }
@@ -481,6 +440,23 @@ numberOfString(const char* a_string,                // string we want to convert
 }
 
 
+template <class NumericType, class Validator1, class Validator2>
+NumericType
+numberOfString(const std::string& a_string,         // string we want to convert into a number
+               Validator1 is_valid1,                // 1st validator function
+               const std::string& invalid_message1, // error message for failed 1st validation
+               NumericType replacement_value1,      // replacement return value on 1st failure
+               Validator2 is_valid2,                // 2nd validator function
+               const std::string& invalid_message2, // error message for failed 2nd validation
+               NumericType replacement_value2)      // replacement return value on 2nd failure
+{
+    numberOfString(&a_string[0],
+                   is_valid1,
+                   invalid_message1, replacement_value1,
+                   is_valid2, invalid_message2, replacement_value2);
+}
+
+
 /**  Convert a_string into a number.
  */
 template <class NumericType, class Validator>
@@ -492,7 +468,26 @@ numberOfString(const char* a_string,               // string we want to convert
 {
     return numberOfString(a_string,
                           is_valid, invalid_message, replacement_value,
-                          boost::lambda::constant(true), "<never reached>", NumericType());
+                          [](NumericType) {return true;},
+                          "<never reached>", NumericType());
+}
+
+
+template <class NumericType, class Validator>
+NumericType
+numberOfString(const std::string& a_string,        // string we want to convert into a number
+               Validator is_valid,                 // validator function
+               const std::string& invalid_message, // error message for failed validation
+               NumericType replacement_value)      // replacement return value on failure
+{
+    return numberOfString(&a_string[0], is_valid, invalid_message, replacement_value);
+}
+
+
+inline bool
+isFloatingPoint(const std::string& aPixelType)
+{
+    return  aPixelType == "FLOAT" || aPixelType == "DOUBLE";
 }
 
 
@@ -502,33 +497,31 @@ std::string
 outputPixelTypeOfString(const char* anOutputDepth)
 {
     typedef std::map<std::string, std::string> Str2StrMapType;
-    Str2StrMapType depthMap;
 
-    boost::assign::insert(depthMap)
-        ("INT16", "INT16")
-        ("INT32", "INT32")
+    Str2StrMapType depthMap = {
+        {"INT16", "INT16"},
+        {"INT32", "INT32"},
 
-        ("8", "UINT8")
-        ("16", "UINT16")
-        ("32", "UINT32")
-        ("UINT8", "UINT8")
-        ("UINT16", "UINT16")
-        ("UINT32", "UINT32")
+        {"8", "UINT8"},
+        {"16", "UINT16"},
+        {"32", "UINT32"},
+        {"UINT8", "UINT8"},
+        {"UINT16", "UINT16"},
+        {"UINT32", "UINT32"},
 
-        ("DOUBLE", "DOUBLE")
-        ("FLOAT", "FLOAT")
-        ("R32", "FLOAT")
-        ("R64", "DOUBLE")
-        ("REAL32", "FLOAT")
-        ("REAL64", "DOUBLE");
+        {"DOUBLE", "DOUBLE"},
+        {"FLOAT", "FLOAT"},
+        {"R32", "FLOAT"},
+        {"R64", "DOUBLE"},
+        {"REAL32", "FLOAT"},
+        {"REAL64", "DOUBLE"}
+    };
 
-    std::string output_depth(anOutputDepth);
-    boost::algorithm::to_upper(output_depth);
+    const std::string output_depth(to_upper_copy(std::string(anOutputDepth)));
     Str2StrMapType::const_iterator p = depthMap.find(output_depth);
     if (p == depthMap.end())
     {
-        throw std::invalid_argument(std::string("unknown output depth \"") +
-                                    anOutputDepth + "\"");
+        throw std::invalid_argument(std::string("unknown output depth \"") + anOutputDepth + "\"");
     }
     else
     {
@@ -543,15 +536,23 @@ std::string
 bestPixelType(const std::string& aFileType, const std::string& aPixelType)
 {
     if (aFileType == "BMP" || aFileType == "JPEG" || aFileType == "RAS")
+    {
         return "UINT8";
+    }
     else if (aFileType == "PNG" &&
              (aPixelType == "INT32" || aPixelType == "UINT32" ||
               aPixelType == "FLOAT" || aPixelType == "DOUBLE"))
+    {
         return "UINT16";
+    }
     else if (aFileType == "EXR")
+    {
         return "FLOAT";
+    }
     else
+    {
         return aPixelType;
+    }
 }
 
 
@@ -563,22 +564,22 @@ range_t
 rangeOfPixelType(const std::string& aPixelType)
 {
     typedef std::map<std::string, range_t> Str2PairMapType;
-    Str2PairMapType rangeMap;
 
-    boost::assign::insert(rangeMap)
-        ("INT8", std::make_pair(vigra::NumericTraits<vigra::Int8>::min(),
-                                vigra::NumericTraits<vigra::Int8>::max()))
-        ("INT16", std::make_pair(vigra::NumericTraits<vigra::Int16>::min(),
-                                 vigra::NumericTraits<vigra::Int16>::max()))
-        ("INT32", std::make_pair(vigra::NumericTraits<vigra::Int32>::min(),
-                                 vigra::NumericTraits<vigra::Int32>::max()))
+    Str2PairMapType rangeMap = {
+        {"INT8", std::make_pair(vigra::NumericTraits<vigra::Int8>::min(),
+                                vigra::NumericTraits<vigra::Int8>::max())},
+        {"INT16", std::make_pair(vigra::NumericTraits<vigra::Int16>::min(),
+                                 vigra::NumericTraits<vigra::Int16>::max())},
+        {"INT32", std::make_pair(vigra::NumericTraits<vigra::Int32>::min(),
+                                 vigra::NumericTraits<vigra::Int32>::max())},
 
-        ("UINT8", std::make_pair(0.0, vigra::NumericTraits<vigra::UInt8>::max()))
-        ("UINT16", std::make_pair(0.0, vigra::NumericTraits<vigra::UInt16>::max()))
-        ("UINT32", std::make_pair(0.0, vigra::NumericTraits<vigra::UInt32>::max()))
+        {"UINT8", std::make_pair(0.0, vigra::NumericTraits<vigra::UInt8>::max())},
+        {"UINT16", std::make_pair(0.0, vigra::NumericTraits<vigra::UInt16>::max())},
+        {"UINT32", std::make_pair(0.0, vigra::NumericTraits<vigra::UInt32>::max())},
 
-        ("FLOAT", std::make_pair(0.0, 1.0))
-        ("DOUBLE", std::make_pair(0.0, 1.0));
+        {"FLOAT", std::make_pair(0.0, 1.0)},
+        {"DOUBLE", std::make_pair(0.0, 1.0)}
+    };
 
     assert(!aPixelType.empty());
     Str2PairMapType::const_iterator r = rangeMap.find(aPixelType);
@@ -625,13 +626,11 @@ maxPixelType(const std::string& aPixelType, const std::string& anotherPixelType)
         return anotherPixelType; // second includes first
     } else {
         // Types are different: look for the smallest containing type
-        using namespace boost::assign;
-
         typedef std::vector<std::string> string_array;
         typedef string_array::const_iterator string_array_ci;
 
         if (range1.first < 0 || range2.first < 0) {
-            const string_array types = list_of("INT8")("INT16")("INT32");
+            const string_array types = {"INT8", "INT16", "INT32"};
             for (string_array_ci i = types.begin(); i != types.end(); ++i) {
                 if (includesBothRanges(*i, range1, range2)) {
                     return *i;
@@ -639,7 +638,7 @@ maxPixelType(const std::string& aPixelType, const std::string& anotherPixelType)
             }
             return "INT32";
         } else {
-            const string_array types = list_of("UINT8")("UINT16")("UINT32");
+            const string_array types = {"UINT8", "UINT16", "UINT32"};
             for (string_array_ci i = types.begin(); i != types.end(); ++i) {
                 if (includesBothRanges(*i, range1, range2)) {
                     return *i;
@@ -744,12 +743,12 @@ expandFilenameTemplate(const std::string& aTemplate,
                     case 'n':
                         ++aNumber;
                         ++aNumberOfImages;
-                        // FALLTHROUGH
+                        BOOST_FALLTHROUGH;
                     case 'i':
                     {
                         std::ostringstream oss;
                         oss <<
-                            std::setw(width.empty() ? 1 + ilog10(aNumberOfImages - 1) : atoi(width.c_str())) <<
+                            std::setw(width.empty() ? 1U + ilog10(aNumberOfImages - 1U) : atoi(width.c_str())) <<
                             std::setfill(pad == 0 ? '0' : pad) <<
                             aNumber;
                         result.append(oss.str());
@@ -799,7 +798,7 @@ expandFilenameTemplate(const std::string& aTemplate,
                         }
                         std::cerr << " in\n"
                                   << command
-                                  << ": warning:     filename template \""
+                                  << ": warning: filename template \""
                                   << aTemplate
                                   << "\""
                                   << std::endl;
@@ -836,13 +835,35 @@ optional_layer_name(unsigned layer_number, unsigned layer_total)
 }
 
 
+template <class predicate>
+inline static void
+trim_if(std::string& a_string, predicate a_predicate)
+{
+    auto negated_predicate =
+        std::bind(std::logical_not<bool>(), std::bind(a_predicate, std::placeholders::_1));
+    std::string::iterator begin = std::find_if(a_string.begin(), a_string.end(), negated_predicate);
+
+    if (begin == a_string.end())
+    {
+        a_string.clear();
+    }
+    else
+    {
+        std::string::reverse_iterator reverse_end =
+            std::find_if(a_string.rbegin(), a_string.rend(), negated_predicate);
+
+        a_string.assign(begin, reverse_end.base());
+    }
+}
+
+
 inline std::string
 profileInfo(cmsHPROFILE profile, cmsInfoType info)
 {
-    const size_t size = cmsGetProfileInfoASCII(profile, info, cmsNoLanguage, cmsNoCountry, NULL, 0);
+    const size_t size = cmsGetProfileInfoASCII(profile, info, cmsNoLanguage, cmsNoCountry, nullptr, 0);
     std::string information(size, '\000');
     cmsGetProfileInfoASCII(profile, info, cmsNoLanguage, cmsNoCountry, &information[0], size);
-    boost::trim_if(information, std::bind2nd(std::less_equal<char>(), '\040'));
+    trim_if(information, std::bind(std::less_equal<char>(), std::placeholders::_1, '\040'));
 
     return information;
 }
@@ -862,242 +883,14 @@ profileName(cmsHPROFILE profile)
 }
 
 
-namespace parameter
+inline unsigned
+profileChannels(cmsHPROFILE profile)
 {
-    // Identifier: [A-Za-z][A-Za-z0-9_-]*
-    inline bool
-    is_valid_identifier(const std::string& identifier)
-    {
-        if (identifier.size() == 0)
-        {
-            return false;
-        }
-        else if (!isalpha(identifier[0]))
-        {
-            return false;
-        }
-        else
-        {
-            for (std::string::const_iterator x = identifier.begin(); x != identifier.end(); ++x)
-            {
-                if (!(isalnum(*x) || *x == '_' || *x == '-'))
-                {
-                    return false;
-                }
-            }
-        }
-
-        return true;
-    }
-
-
-    struct not_found : public std::runtime_error
-    {
-        not_found(const std::string& a_message) : std::runtime_error(a_message) {}
-    };
-
-
-    // Notes:
-    //
-    // * The access of parameters through enblend::parameter::as_* is
-    //   reasonable fast.  For time-critical parts of the code, the
-    //   parameter's value can always be copied into a local variable.
-    //
-    // * The map from parameter keys to values is meant to be constant
-    //   after the command line was parsed, i.e. neither the map
-    //   itself, nor one of its values should be modified.  Following
-    //   this convention makes all enblend::parameter::as_* functions
-    //   thread safe.
-    //
-    // Some examples how to use parameters:
-    //
-    // (1) Check whether a parameter has been set.
-    //         if (enblend::parameter::exists("foobar")) {...}
-    //         else {...}
-    //
-    // (2) Use a parameter that is known to exist.
-    //         std::string s = enblend::parameter::as_string("foobar");
-    //         int i = enblend::parameter::as_integer("foobar");
-    //         unsigned u = enblend::parameter::as_unsigned("foobar");
-    //         double x = enblend::parameter::as_floating_point("foobar");
-    //         bool b = enblend::parameter::as_boolean("foobar");
-    //
-    // (3) Substitute parameter value if it exists; otherwise go with
-    //     the default.
-    //         std::string s = enblend::parameter::as_string("foobar", "baz");
-    //         int i = enblend::parameter::as_integer("foobar", 123);
-    //         unsigned u = enblend::parameter::as_unsigned("foobar", 42U);
-    //         double x = enblend::parameter::as_floating_point("foobar", 0.577215665);
-    //         bool b = enblend::parameter::as_boolean("foobar", true);
-    //
-    // (4) React on parameter with a non-local change of control flow
-    //         int i;
-    //         try {i = enblend::parameter::as_integer("foobar");}
-    //         catch (enblend::parameter::not_found&) {...}
-    //
-    // A parameter always can be retrieved as string with function
-    // as_string().  All other as_* functions throw the exception
-    // conversion_error, if the parameter's value cannot be
-    // represented.  See class ParameterValue in file "global.h" for
-    // the details on conversion especially for as_boolean().
-
-    inline bool
-    exists(const std::string& key)
-    {
-        return Parameter.find(key) != Parameter.end();
-    }
-
-
-    inline std::string
-    as_string(const std::string& key)
-    {
-        parameter_map::const_iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            throw not_found(key);
-        }
-        else
-        {
-            return x->second.as_string();
-        }
-    }
-
-
-    inline std::string
-    as_string(const std::string& key, const std::string& default_value)
-    {
-        parameter_map::const_iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            return default_value;
-        }
-        else
-        {
-            return x->second.as_string();
-        }
-    }
-
-
-    inline int
-    as_integer(const std::string& key)
-    {
-        parameter_map::iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            throw not_found(key);
-        }
-        else
-        {
-            return x->second.as_integer();
-        }
-    }
-
-
-    inline int
-    as_integer(const std::string& key, int default_value)
-    {
-        parameter_map::iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            return default_value;
-        }
-        else
-        {
-            return x->second.as_integer();
-        }
-    }
-
-
-    inline unsigned
-    as_unsigned(const std::string& key)
-    {
-        parameter_map::iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            throw not_found(key);
-        }
-        else
-        {
-            return x->second.as_unsigned();
-        }
-    }
-
-
-    inline unsigned
-    as_unsigned(const std::string& key, unsigned default_value)
-    {
-        parameter_map::iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            return default_value;
-        }
-        else
-        {
-            return x->second.as_unsigned();
-        }
-    }
-
-
-    inline double
-    as_double(const std::string& key)
-    {
-        parameter_map::iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            throw not_found(key);
-        }
-        else
-        {
-            return x->second.as_double();
-        }
-    }
-
-
-    inline double
-    as_double(const std::string& key, double default_value)
-    {
-        parameter_map::iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            return default_value;
-        }
-        else
-        {
-            return x->second.as_double();
-        }
-    }
-
-
-    inline bool
-    as_boolean(const std::string& key)
-    {
-        parameter_map::iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            throw not_found(key);
-        }
-        else
-        {
-            return x->second.as_boolean();
-        }
-    }
-
-
-    inline bool
-    as_boolean(const std::string& key, bool default_value)
-    {
-        parameter_map::iterator x = Parameter.find(key);
-        if (x == Parameter.end())
-        {
-            return default_value;
-        }
-        else
-        {
-            return x->second.as_boolean();
-        }
-    }
-} // namespace parameter
-
+    const cmsColorSpaceSignature signature = cmsGetColorSpace(profile);
+    const unsigned number_of_channels = cmsChannelsOf(signature);
+    assert(number_of_channels == 1U || number_of_channels == 3U);
+    return number_of_channels;
+}
 } // namespace enblend
 
 
@@ -1133,12 +926,12 @@ strtok_r(char* str, const char* delim, char** save_ptr)
         }
         else
         {
-            return NULL;
+            return nullptr;
         }
     }
     else
     {
-        return NULL;
+        return nullptr;
     }
 }
 #endif
