@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 2004-2012 Andrew Mihal
+ * Copyright (C) 2004-2009 Andrew Mihal
+ * Copyright (C) 2009-2016 Christoph Spiel
  *
  * This file is part of Enblend.
  *
@@ -27,16 +28,16 @@
 #include <iostream>
 #include <list>
 
-#include <boost/static_assert.hpp>
-
 #include <vigra/impex.hxx>
 #include <vigra/initimage.hxx>
 #include <vigra/transformimage.hxx>
 
-#include "vigra_ext/rect2d.hxx"
+#include "rect2d.hxx"
 
 #include "common.h"
-#include "openmp.h"
+#include "opencl.h"
+#include "openmp_def.h"
+#include "openmp_vigra.h"
 #include "numerictraits.h"
 #include "fixmath.h"
 #include "assemble.h"
@@ -91,18 +92,6 @@ void enblendMain(const FileNameList& anInputFileNameList,
     //                !OneAtATime: 2*anInputUnion*imageValueType + 2*anInputUnion*AlphaValueType
     // mem usage after = anInputUnion*ImageValueType + anInputUnion*AlphaValueType
 
-#ifdef CACHE_IMAGES
-    if (Verbose >= VERBOSE_CFI_MESSAGES) {
-        vigra_ext::CachedFileImageDirector& v = vigra_ext::CachedFileImageDirector::v();
-        std::cerr << command
-                  << ": info: image cache statistics after loading black image\n";
-        v.printStats(std::cerr, command + ": info:     blackImage", blackPair.first);
-        v.printStats(std::cerr, command + ": info:     blackAlpha", blackPair.second);
-        v.printStats(std::cerr, command + ": info: ");
-        v.resetCacheMisses();
-    }
-#endif
-
     const unsigned numberOfImages = imageInfoList.size();
 
     // Main blending loop.
@@ -119,20 +108,6 @@ void enblendMain(const FileNameList& anInputFileNameList,
         //                !OneAtATime: 2*anInputUnion*imageValueType + 2*anInputUnion*AlphaValueType
         // mem usage after = 2*anInputUnion*ImageValueType + 2*anInputUnion*AlphaValueType
 
-#ifdef CACHE_IMAGES
-        if (Verbose >= VERBOSE_CFI_MESSAGES) {
-            vigra_ext::CachedFileImageDirector& v = vigra_ext::CachedFileImageDirector::v();
-            std::cerr << command
-                      <<": info: image cache statistics after loading white image\n";
-            v.printStats(std::cerr, command + ": info:     blackImage", blackPair.first);
-            v.printStats(std::cerr, command + ": info:     blackAlpha", blackPair.second);
-            v.printStats(std::cerr, command + ": info:     whiteImage", whitePair.first);
-            v.printStats(std::cerr, command + ": info:     whiteAlpha", whitePair.second);
-            v.printStats(std::cerr, command + ": info: ");
-            v.resetCacheMisses();
-        }
-#endif
-
         // Union bounding box of whiteImage and blackImage.
         vigra::Rect2D uBB = blackBB | whiteBB;
 
@@ -167,22 +142,39 @@ void enblendMain(const FileNameList& anInputFileNameList,
             // White image is redundant.
             delete whitePair.first;
             delete whitePair.second;
-            std::cerr << command << ": warning: some images are redundant and will not be blended"
-                      << std::endl;
+            std::cerr << command << ": warning: some images are redundant and will not be blended\n"
+                      << command << ": note: usually this means that at least one of the images\n"
+                      << command << ": note: does not belong to the set" << std::endl;
             continue;
         } else if (overlap == NoOverlap && ExactLevels == 0) {
             // Images do not actually overlap.
-            std::cerr << command << ": images do not overlap - they will be combined without blending\n"
-                      << command << ": use the \"-l\" flag to force blending with a certain number of levels"
-                      << std::endl;
+            std::cerr << command << ": warning: images do not overlap; they will be combined without blending\n"
+                      << command << ": note: usually this means that at least one of these images does not\n"
+                      << command << ": note: belong to the set or the order of the images given at the\n"
+                      << command << ": note: command line is wrong; sometimes passing option\n"
+                      << command << ": note: \"--pre-assemble\" resolves the problem; in rare cases using\n"
+                      << command << ": note: option \"--levels=NUMBER\" to force blending with a certain\n"
+                      << command << ": note: NUMBER of levels can help, too" << std::endl;
 
             // Copy white image into black image verbatim.
-            vigra::copyImageIf(srcImageRange(*(whitePair.first)),
-                               maskImage(*(whitePair.second)),
-                               destImage(*(blackPair.first)));
-            vigra::copyImageIf(srcImageRange(*(whitePair.second)),
-                               maskImage(*(whitePair.second)),
-                               destImage(*(blackPair.second)));
+#ifdef OPENMP
+            omp::scoped_nested(true);
+            omp::scoped_dynamic(true);
+#pragma omp parallel
+            {
+#pragma omp single nowait
+                {
+#endif
+                    vigra::omp::copyImageIf(srcImageRange(*(whitePair.first)),
+                                            maskImage(*(whitePair.second)),
+                                            destImage(*(blackPair.first)));
+                    vigra::omp::copyImageIf(srcImageRange(*(whitePair.second)),
+                                            maskImage(*(whitePair.second)),
+                                            destImage(*(blackPair.second)));
+#ifdef OPENMP
+                } // omp single
+            } // omp parallel
+#endif
 
             delete whitePair.first;
             delete whitePair.second;
@@ -307,21 +299,6 @@ void enblendMain(const FileNameList& anInputFileNameList,
         //                  2*anInputUnion*ImageValueType +
         //                  2*anInputUnion*AlphaValueType
 
-#ifdef CACHE_IMAGES
-        if (Verbose >= VERBOSE_CFI_MESSAGES) {
-            vigra_ext::CachedFileImageDirector& v = vigra_ext::CachedFileImageDirector::v();
-            std::cerr << command
-                      << ": info: image cache statistics after mask generation\n";
-            v.printStats(std::cerr, command + ": info:     blackImage", blackPair.first);
-            v.printStats(std::cerr, command + ": info:     blackAlpha", blackPair.second);
-            v.printStats(std::cerr, command + ": info:     whiteImage", whitePair.first);
-            v.printStats(std::cerr, command + ": info:     whiteAlpha", whitePair.second);
-            v.printStats(std::cerr, command + ": info:     mask", mask);
-            v.printStats(std::cerr, command + ": info: ");
-            v.resetCacheMisses();
-        }
-#endif
-
         // Calculate ROI bounds and number of levels from mBB.
         // ROI bounds must be at least mBB but not to extend uBB.
         vigra::Rect2D roiBB;
@@ -392,24 +369,6 @@ void enblendMain(const FileNameList& anInputFileNameList,
         // mem usage after = MaskType*ubb + 2*anInputUnion*ImageValueType + 2*anInputUnion*AlphaValueType
         //                   + (4/3)*roiBB*MaskPyramidType
 
-#ifdef CACHE_IMAGES
-        if (Verbose >= VERBOSE_CFI_MESSAGES) {
-            vigra_ext::CachedFileImageDirector& v = vigra_ext::CachedFileImageDirector::v();
-            std::cerr << command
-                      << ": info: image cache statistics after calculating mask pyramid\n";
-            v.printStats(std::cerr, command + ": info:     blackImage", blackPair.first);
-            v.printStats(std::cerr, command + ": info:     blackAlpha", blackPair.second);
-            v.printStats(std::cerr, command + ": info:     whiteImage", whitePair.first);
-            v.printStats(std::cerr, command + ": info:     whiteAlpha", whitePair.second);
-            v.printStats(std::cerr, command + ": info:     mask", mask);
-            for (unsigned int i = 0; i < maskGP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     maskGP", i, (*maskGP)[i]);
-            }
-            v.printStats(std::cerr, command + ": info: ");
-            v.resetCacheMisses();
-        }
-#endif
-
         // Now it is safe to make changes to mask image.
         // Black out the ROI in the mask.
         // Make an roiBounds relative to uBB origin.
@@ -434,30 +393,12 @@ void enblendMain(const FileNameList& anInputFileNameList,
         std::vector<ImagePyramidType*>* whiteLP =
             laplacianPyramid<ImageType, AlphaType, ImagePyramidType,
                              ImagePyramidIntegerBits, ImagePyramidFractionBits,
-                             SKIPSMImagePixelType, SKIPSMAlphaPixelType>("whiteGP",
-                                                                         numLevels, wraparoundForBlend,
-                                                                         vigra_ext::apply(roiBB, srcImageRange(*(whitePair.first))),
-                                                                         vigra_ext::apply(roiBB, maskImage(*(whitePair.second))));
+                             SKIPSMImagePixelType, SKIPSMAlphaPixelType>
+            ("whiteGP",
+             numLevels, wraparoundForBlend,
+             vigra_ext::apply(roiBB, srcImageRange(*(whitePair.first))),
+             vigra_ext::apply(roiBB, maskImage(*(whitePair.second))));
 
-#ifdef CACHE_IMAGES
-        if (Verbose >= VERBOSE_CFI_MESSAGES) {
-            vigra_ext::CachedFileImageDirector& v = vigra_ext::CachedFileImageDirector::v();
-            std::cerr << command
-                      << ": info: image cache statistics after calculating white pyramid\n";
-            v.printStats(std::cerr, command + ": info:     blackImage", blackPair.first);
-            v.printStats(std::cerr, command + ": info:     blackAlpha", blackPair.second);
-            v.printStats(std::cerr, command + ": info:     whiteImage", whitePair.first);
-            v.printStats(std::cerr, command + ": info:     whiteAlpha", whitePair.second);
-            for (unsigned int i = 0; i < maskGP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     maskGP", i, (*maskGP)[i]);
-            }
-            for (unsigned int i = 0; i < whiteLP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     whiteLP", i, (*whiteLP)[i]);
-            }
-            v.printStats(std::cerr, command + ": info: ");
-            v.resetCacheMisses();
-        }
-#endif
         // mem usage after = 2*anInputUnion*ImageValueType + 2*anInputUnion*AlphaValueType
         //                   + (4/3)*roiBB*MaskPyramidType + (4/3)*roiBB*ImagePyramidType
         // mem xsection = 4 * roiBB.width() * SKIPSMImagePixelType
@@ -472,32 +413,11 @@ void enblendMain(const FileNameList& anInputFileNameList,
         std::vector<ImagePyramidType*>* blackLP =
             laplacianPyramid<ImageType, AlphaType, ImagePyramidType,
                              ImagePyramidIntegerBits, ImagePyramidFractionBits,
-                             SKIPSMImagePixelType, SKIPSMAlphaPixelType>("blackGP",
-                                                                         numLevels, wraparoundForBlend,
-                                                                         vigra_ext::apply(roiBB, srcImageRange(*(blackPair.first))),
-                                                                         vigra_ext::apply(roiBB, maskImage(*(blackPair.second))));
-
-#ifdef CACHE_IMAGES
-        if (Verbose >= VERBOSE_CFI_MESSAGES) {
-            vigra_ext::CachedFileImageDirector& v = vigra_ext::CachedFileImageDirector::v();
-            std::cerr << command
-                      << ": info: image cache statistics after calculating black pyramid\n";
-            v.printStats(std::cerr, command + ": info:     blackImage", blackPair.first);
-            v.printStats(std::cerr, command + ": info:     blackAlpha", blackPair.second);
-            v.printStats(std::cerr, command + ": info:     whiteAlpha", whitePair.second);
-            for (unsigned int i = 0; i < maskGP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     maskGP", i, (*maskGP)[i]);
-            }
-            for (unsigned int i = 0; i < whiteLP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     whiteLP", i, (*whiteLP)[i]);
-            }
-            for (unsigned int i = 0; i < blackLP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     blackLP", i, (*blackLP)[i]);
-            }
-            v.printStats(std::cerr, command + ": info: ");
-            v.resetCacheMisses();
-        }
-#endif
+                             SKIPSMImagePixelType, SKIPSMAlphaPixelType>
+            ("blackGP",
+             numLevels, wraparoundForBlend,
+             vigra_ext::apply(roiBB, srcImageRange(*(blackPair.first))),
+             vigra_ext::apply(roiBB, maskImage(*(blackPair.second))));
 
 #ifdef DEBUG_EXPORT_PYRAMID
         exportPyramid<SKIPSMImagePixelType, ImagePyramidType>(blackLP, "enblend_black_lp");
@@ -523,28 +443,9 @@ void enblendMain(const FileNameList& anInputFileNameList,
         //      + (4/3)*roiBB*MaskPyramidType + 2*(4/3)*roiBB*ImagePyramidType
 
         // Blend pyramids
-        ConvertScalarToPyramidFunctor<MaskPixelType, MaskPyramidPixelType, MaskPyramidIntegerBits, MaskPyramidFractionBits> whiteMask;
+        ConvertScalarToPyramidFunctor<MaskPixelType, MaskPyramidPixelType,
+                                      MaskPyramidIntegerBits, MaskPyramidFractionBits> whiteMask;
         blend(maskGP, whiteLP, blackLP, whiteMask(vigra::NumericTraits<MaskPixelType>::max()));
-#ifdef CACHE_IMAGES
-        if (Verbose >= VERBOSE_CFI_MESSAGES) {
-            vigra_ext::CachedFileImageDirector& v = vigra_ext::CachedFileImageDirector::v();
-            std::cerr << command
-                      << ": info: image cache statistics after blending pyramids\n";
-            v.printStats(std::cerr, command + ": info:     blackImage", blackPair.first);
-            v.printStats(std::cerr, command + ": info:     blackAlpha", blackPair.second);
-            for (unsigned int i = 0; i < maskGP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     maskGP", i, (*maskGP)[i]);
-            }
-            for (unsigned int i = 0; i < whiteLP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     whiteLP", i, (*whiteLP)[i]);
-            }
-            for (unsigned int i = 0; i < blackLP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     blackLP", i, (*blackLP)[i]);
-            }
-            v.printStats(std::cerr, command + ": info: ");
-            v.resetCacheMisses();
-        }
-#endif
 
         // delete mask pyramid
 #ifdef DEBUG_EXPORT_PYRAMID
@@ -574,26 +475,13 @@ void enblendMain(const FileNameList& anInputFileNameList,
 
         // collapse black pyramid
         collapsePyramid<SKIPSMImagePixelType>(wraparoundForBlend, blackLP);
-#ifdef CACHE_IMAGES
-        if (Verbose >= VERBOSE_CFI_MESSAGES) {
-            vigra_ext::CachedFileImageDirector& v = vigra_ext::CachedFileImageDirector::v();
-            std::cerr << command
-                      << ": info: image cache statistics after collapsing black pyramid\n";
-            v.printStats(std::cerr, command + ": info:     blackImage", blackPair.first);
-            v.printStats(std::cerr, command + ": info:     blackAlpha", blackPair.second);
-            for (unsigned int i = 0; i < blackLP->size(); i++) {
-                v.printStats(std::cerr, command + ": info:     blackLP", i, (*blackLP)[i]);
-            }
-            v.printStats(std::cerr, command + ": info: ");
-            v.resetCacheMisses();
-        }
-#endif
 
         // copy collapsed black pyramid into black image ROI, using black alpha mask.
         copyFromPyramidImageIf<ImagePyramidType, MaskType, ImageType,
-                               ImagePyramidIntegerBits, ImagePyramidFractionBits>(srcImageRange(*((*blackLP)[0])),
-                                                                                  vigra_ext::apply(roiBB, maskImage(*(blackPair.second))),
-                                                                                  vigra_ext::apply(roiBB, destImage(*(blackPair.first))));
+                               ImagePyramidIntegerBits, ImagePyramidFractionBits>
+            (srcImageRange(*((*blackLP)[0])),
+             vigra_ext::apply(roiBB, maskImage(*(blackPair.second))),
+             vigra_ext::apply(roiBB, destImage(*(blackPair.first))));
 
         // delete black pyramid
         for (unsigned int i = 0; i < blackLP->size(); i++) {
@@ -616,18 +504,6 @@ void enblendMain(const FileNameList& anInputFileNameList,
             checkpoint(blackPair, anOutputImageInfo);
         }
 
-#ifdef CACHE_IMAGES
-        if (Verbose >= VERBOSE_CFI_MESSAGES) {
-            vigra_ext::CachedFileImageDirector& v = vigra_ext::CachedFileImageDirector::v();
-            std::cerr << command
-                      << ": info: image cache statistics after checkpointing\n";
-            v.printStats(std::cerr, command + ": info:     blackImage", blackPair.first);
-            v.printStats(std::cerr, command + ": info:     blackAlpha", blackPair.second);
-            v.printStats(std::cerr, command + ": info: ");
-            v.resetCacheMisses();
-        }
-#endif
-
         // Now set blackBB to uBB.
         blackBB = uBB;
 
