@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2015 Christoph L. Spiel
+ * Copyright (C) 2009-2016 Christoph L. Spiel
  *
  * This file is part of Enblend.
  *
@@ -28,6 +28,7 @@
 #include <unistd.h>
 #endif
 
+#include <cctype>
 #include <cerrno>
 #include <fstream>
 #include <iomanip>
@@ -36,8 +37,6 @@
 #include <map>
 #include <cctype>
 
-#include <boost/algorithm/string/case_conv.hpp>
-
 #include <vigra/imageinfo.hxx>
 
 #ifdef HAVE_CONFIG_H
@@ -58,7 +57,7 @@
 
 
 // Separator of key and value in a syntactic comment
-#define KEY_VALUE_SEPARATOR_CHAR ':'
+#define KEY_VALUE_SEPARATOR_CHAR ':' //< response-file-syntactic-comment-key-value-separator \char58
 
 
 typedef std::pair<std::string, std::string> key_value_pair;
@@ -71,6 +70,102 @@ extern LayerSelectionHost LayerSelection;
 
 namespace enblend
 {
+// Print the (back-)trace of all response files opened so far.
+extern void unroll_trace(const FilePositionTrace& a_trace);
+
+
+TraceableFileName::TraceableFileName(const std::string& a_filename,
+                                     const FilePositionTrace& a_trace,
+                                     selector::Abstract* a_selector) :
+    selector_(a_selector), filename_(a_filename), trace_(a_trace)
+{}
+
+
+TraceableFileName::~TraceableFileName()
+{}
+
+
+TraceableFileName* TraceableFileName::clone() const
+{
+    return new TraceableFileName(*this);
+}
+
+
+const std::string&
+TraceableFileName::filename() const
+{
+    return filename_;
+}
+
+
+const FilePositionTrace&
+TraceableFileName::trace() const
+{
+    return trace_;
+}
+
+
+void
+TraceableFileName::unroll_trace() const
+{
+    enblend::unroll_trace(trace_);
+}
+
+
+selector::Abstract*
+TraceableFileName::selector() const
+{
+    return selector_;
+}
+
+
+TraceableFileNameAndLayer::TraceableFileNameAndLayer(const std::string& a_filename,
+                                                     const FilePositionTrace& a_trace,
+                                                     const std::string& a_layer_specification) :
+    super(a_filename, a_trace, new selector::IndexedLayer(a_layer_specification)),
+    layer_specification_(a_layer_specification)
+{
+#ifdef DEBUG_FILESPEC
+    std::cout <<
+        "+ TraceableFileNameAndLayer(,,,): filename = <" << filename() << ">\n" <<
+        "+ TraceableFileNameAndLayer(,,,): layer_spec = <" << layer_spec() << ">\n" <<
+        "+ TraceableFileNameAndLayer(,,,): selector = <" << selector()->name() << ">\n";
+#endif
+}
+
+
+TraceableFileNameAndLayer::TraceableFileNameAndLayer(const TraceableFileNameAndLayer& another) :
+    super(another.filename(), another.trace(), new selector::IndexedLayer(another.layer_spec())),
+    layer_specification_(another.layer_specification_)
+{
+#ifdef DEBUG_FILESPEC
+    std::cout <<
+        "+ TraceableFileNameAndLayer(const&): filename = <" << filename() << ">\n" <<
+        "+ TraceableFileNameAndLayer(const&): layer_spec = <" << layer_spec() << ">\n" <<
+        "+ TraceableFileNameAndLayer(const&): selector = <" << selector()->name() << ">\n";
+#endif
+}
+
+
+TraceableFileNameAndLayer::~TraceableFileNameAndLayer()
+{
+    delete selector_;
+}
+
+
+TraceableFileNameAndLayer*
+TraceableFileNameAndLayer::clone() const
+{
+    return new TraceableFileNameAndLayer(*this);
+}
+
+
+std::string
+TraceableFileNameAndLayer::layer_spec() const
+{
+    return layer_specification_;
+}
+
 
 #ifdef _WIN32
 /** Add all files which match filename to filelist.  filename can
@@ -86,7 +181,7 @@ glob_filename_win32(FileNameList& filelist, const std::string& filename)
     char ext[_MAX_EXT];
     char newFile[_MAX_PATH];
 
-    _splitpath(filename.c_str(), drive, dir, NULL, NULL);
+    _splitpath(filename.c_str(), drive, dir, nullptr, nullptr);
 
     struct _finddata_t finddata;
     intptr_t findhandle = _findfirst(filename.c_str(), &finddata);
@@ -94,7 +189,7 @@ glob_filename_win32(FileNameList& filelist, const std::string& filename)
     {
         do
         {
-            _splitpath(finddata.name, NULL, NULL, fname, ext);
+            _splitpath(finddata.name, nullptr, nullptr, fname, ext);
             _makepath(newFile, drive, dir, fname, ext);
             filelist.push_back(std::string(newFile));
         }
@@ -134,7 +229,7 @@ can_open_file(const std::string& aFilename)
 }
 
 
-/** Convert a_string in a string that is printable.  This is it only
+/** Convert a_string in a string that is printable.  This is, it only
  *  contains printable characters. */
 std::string
 printable_string(const std::string& a_string)
@@ -258,7 +353,7 @@ unroll_trace(const FilePositionTrace& trace)
     for (FilePositionTrace::const_iterator t = trace.begin(); t != trace.end(); ++t)
     {
         std::cerr << command <<
-            ": info:    included from \"" << t->first << "\" line " << t->second << '\n';
+            ": info: included from \"" << t->first << "\" line " << t->second << '\n';
     }
 }
 
@@ -326,7 +421,7 @@ public:
         errno = 0;
         if (glob(a_filespec.c_str(),
                  flags,
-                 NULL, // (*errfunc)(const char* filename, int error_code)
+                 nullptr, // (*errfunc)(const char* filename, int error_code)
                  result_vector_) != 0)
         {
             std::cerr << command <<
@@ -340,7 +435,7 @@ public:
     {
         result_.clear();
 
-        for (char** path = result_vector_->gl_pathv; *path != NULL; ++path)
+        for (char** path = result_vector_->gl_pathv; *path != nullptr; ++path)
         {
             result_.push_back(*path);
         }
@@ -403,20 +498,22 @@ public:
 class Globbing
 {
     // Map from the names of the algorithms or aliases to the
-    // algorithm itself.  The boolean flag indicates whether we point
-    // to the algorithm proper or the name is just an alias to an
-    // existing algorithm.
+    // algorithms themselves.  The boolean flag indicates whether we
+    // point to the algorithm proper or the name is just an alias to
+    // an existing algorithm.
     typedef std::map<std::string, std::pair<bool, AbstractGlobbingAlgorithm*> > algorithm_map;
 
 public:
-    Globbing() : algorithm_name_("literal"), algorithm_(NULL)
+    Globbing() : algorithm_name_("literal"), algorithm_(nullptr)
     {
         installed_algorithms_["literal"] = MAKE_ALGORITHM(new LiteralGlobbingAlgorithm);
         installed_algorithms_["wildcard"] = MAKE_ALGORITHM(new WildcardGlobbingAlgorithm);
-        setup_alias("literal", "none");
-
 #ifndef _WIN32
         installed_algorithms_["shell"] = MAKE_ALGORITHM(new ShellGlobbingAlgorithm);
+#endif
+
+        setup_alias("literal", "none");
+#ifndef _WIN32
         setup_alias("shell", "sh");
 #endif
     }
@@ -433,7 +530,7 @@ public:
             }
         }
 
-        algorithm_ = NULL;
+        algorithm_ = nullptr;
     }
 
     const std::string& get_algorithm() const
@@ -458,7 +555,7 @@ public:
 
     FileNameList expand(const std::string& a_filename, const FilePositionTrace& trace)
     {
-        if (algorithm_ == NULL)
+        if (algorithm_ == nullptr)
         {
             set_algorithm(algorithm_name_);
         }
@@ -504,7 +601,7 @@ known_globbing_algorithms()
 }
 
 
-//< src::RESPONSE_FILE_MAX_NESTING_LEVEL 63
+//< RESPONSE_FILE_MAX_NESTING_LEVEL 63
 #define RESPONSE_FILE_MAX_NESTING_LEVEL 63U
 
 
@@ -523,9 +620,104 @@ struct TraceInfo
 };
 
 
-void
-unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const std::string& filename)
+#define LAYERSPEC_OPEN  "["     //< LAYERSPEC_OPEN \char91
+#define LAYERSPEC_CLOSE "]"     //< LAYERSPEC_CLOSE \char93
+
+
+#ifdef DEBUG_FILESPEC
+static std::string
+pretty_string_of_string_size_type(std::string::size_type a_size)
 {
+    std::ostringstream stream;
+
+    if (a_size == std::string::npos)
+    {
+        stream << "n/pos";
+    }
+    else
+    {
+        stream << a_size;
+    }
+
+    return stream.str();
+}
+#endif
+
+
+bool
+separate_filename_and_optional_layerspec(std::string& filename, std::string& layerspec,
+                                         const std::string& filename_and_optional_layerspec)
+{
+    const std::string::size_type layer_index_begin =
+        filename_and_optional_layerspec.find_last_of(LAYERSPEC_OPEN);
+    const std::string::size_type layer_index_end =
+        filename_and_optional_layerspec.find_last_of(LAYERSPEC_CLOSE);
+    bool has_layerspec;
+
+    if (layer_index_begin != std::string::npos && layer_index_end != std::string::npos &&
+        layer_index_end + 1 == filename_and_optional_layerspec.length())
+    {
+        std::string::size_type filename_end(layer_index_begin - 1);
+        // Allow for whitespace between the filename proper and begin
+        // of the later specification (LAYERSPEC_OPEN).
+        while (std::isspace(filename_and_optional_layerspec[filename_end]))
+        {
+            --filename_end;
+        }
+        filename.assign(filename_and_optional_layerspec, 0, filename_end + 1);
+        layerspec.assign(filename_and_optional_layerspec,
+                         layer_index_begin + 1,
+                         layer_index_end - layer_index_begin - 1);
+        has_layerspec = true;
+    }
+    else
+    {
+        filename.assign(filename_and_optional_layerspec);
+        layerspec.clear();
+        has_layerspec = false;
+    }
+
+#ifdef DEBUG_FILESPEC
+    std::cout <<
+        "+ separate_filename_and_optional_layerspec: layer_index_begin = " <<
+        pretty_string_of_string_size_type(layer_index_begin) <<
+        ", layer_index_end = " <<
+        pretty_string_of_string_size_type(layer_index_end) <<
+        ", length() = " << filename_and_optional_layerspec.length() << "\n" <<
+        "+ separate_filename_and_optional_layerspec: filename = <" << filename <<
+        ">, layerspec = <" << layerspec << ">\n" <<
+        "+ separate_filename_and_optional_layerspec -> " << has_layerspec << "\n";
+#endif
+
+    return has_layerspec;
+}
+
+
+static enblend::TraceableFileName*
+make_traceable_file_name_and_maybe_layer(const std::string& a_filename,
+                                         const FilePositionTrace& a_file_position_trace,
+                                         selector::Abstract* a_selector,
+                                         bool has_layerspec,
+                                         const std::string& a_layer_specification)
+{
+    // We can have an empty layerspec ("[]"): has_layerspec == true && layerspec == "",
+    // in contrast, without layerspec we get: has_layerspec == false && layerspec == "".
+    return
+        has_layerspec ?
+        new enblend::TraceableFileNameAndLayer(a_filename, a_file_position_trace, a_layer_specification) :
+        new enblend::TraceableFileName(a_filename, a_file_position_trace, a_selector);
+}
+
+
+void
+unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info,
+                     const std::string& filename_and_optional_layerspec)
+{
+    std::string filename;
+    std::string layerspec;
+    const bool has_layerspec = separate_filename_and_optional_layerspec(filename, layerspec,
+                                                                        filename_and_optional_layerspec);
+
     // Checking the nesting_lavel acts as an emergency break in the
     // case our usual recursion detection fails.
     if (trace_info.nesting_level > RESPONSE_FILE_MAX_NESTING_LEVEL)
@@ -533,8 +725,9 @@ unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const
         std::cerr <<
             command << ": warning: excessive nesting of " << trace_info.nesting_level <<
             " levels of response files;\n" <<
-            command << ": warning:     possible infinite recursion in \"" <<
-            printable_string(filename) << "\"\n";
+            command << ": warning: possible infinite recursion in \"" <<
+            printable_string(filename) << "\";\n" <<
+            command << ": warning: skipping further files\n";
         unroll_trace(trace_info.file_position);
         return;
     }
@@ -544,6 +737,10 @@ unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const
     if (filename.empty())
     {
         std::cerr << command << ": info: empty filename\n";
+        if (!layerspec.empty())
+        {
+            std::cerr << command << ": info: found only layer specification, missing filename\n";
+        }
         unroll_trace(trace_info.file_position);
         return;
     }
@@ -625,7 +822,7 @@ unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const
                     std::cerr <<
                         command << ": warning: ignoring last line of response file \"" <<
                         printable_string(response_filepath) << "\" line " << line_number + 1U << ",\n" <<
-                        command << ": warning:     because it does not end with a newline\n";
+                        command << ": warning: because it does not end with a newline\n";
                     unroll_trace(trace_info.file_position);
                 }
                 break;
@@ -633,8 +830,8 @@ unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const
             ++line_number;
 
             key_value_pair comment = get_syntactic_comment(buffer);
-            boost::algorithm::to_lower(comment.first);
-            boost::algorithm::to_lower(comment.second);
+            enblend::to_lower(comment.first);
+            enblend::to_lower(comment.second);
             if ((comment.first == "glob" ||
                  comment.first == "globbing" ||
                  comment.first == "filename-globbing") &&
@@ -649,7 +846,7 @@ unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const
                         "\" in response file \"" << printable_string(response_filepath) <<
                         "\" line " << line_number << "\n" <<
                         command <<
-                        ": warning:     will stick with algorithm \"" <<
+                        ": note: will stick with algorithm \"" <<
                         glob.get_algorithm() << "\"\n";
                     unroll_trace(trace_info.file_position);
                 }
@@ -671,13 +868,13 @@ unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const
                         "\" in response file \"" << printable_string(response_filepath) <<
                         "\" line " << line_number << "\n" <<
                         command <<
-                        ": warning:     will stick with selector \"" <<
+                        ": note: will stick with selector \"" <<
                         LayerSelection.name() << "\"\n";
                     unroll_trace(trace_info.file_position);
                 }
                 else
                 {
-                    selector = *new_selector;
+                    selector = new_selector->get();
 #ifdef DEBUG_FILESPEC
                     std::cout << "+ unfold_filename_iter: new layer selection algorithm = " <<
                         selector->name() << "\n";
@@ -707,7 +904,7 @@ unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const
                  ++p)
             {
                 const FileNameList expanded_partial_result =
-                    glob.expand(p->filename(), trace_info.file_position);
+                    glob.expand((*p)->filename(), trace_info.file_position);
                 for (FileNameList::const_iterator q = expanded_partial_result.begin();
                      q != expanded_partial_result.end();
                      ++q)
@@ -719,25 +916,34 @@ unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const
 #ifdef DEBUG_FILESPEC
                         std::cout <<
                             "+ unfold_filename_iter: relative path\n" <<
-                            "+ unfold_filename_iter:     q = " << *q << "\n" <<
-                            "+ unfold_filename_iter:     path = <" << path << ">\n";
+                            "+ unfold_filename_iter: q = " << *q << "\n" <<
+                            "+ unfold_filename_iter: path = <" << path << ">\n";
 #endif
                         result.insert(result.end(),
-                                      enblend::TraceableFileName(path, p->trace(), p->selector()));
+                                      make_traceable_file_name_and_maybe_layer(path, (*p)->trace(),
+                                                                               (*p)->selector(),
+                                                                               has_layerspec, layerspec));
                     }
                     else
                     {
 #ifdef DEBUG_FILESPEC
                         std::cout <<
                             "+ unfold_filename_iter: absolute path\n" <<
-                            "+ unfold_filename_iter:     q = " << *q << "\n";
+                            "+ unfold_filename_iter: q = " << *q << "\n";
 #endif
                         result.insert(result.end(),
-                                      enblend::TraceableFileName(*q, p->trace(), p->selector()));
+                                      make_traceable_file_name_and_maybe_layer(*q, (*p)->trace(),
+                                                                               (*p)->selector(),
+                                                                               has_layerspec, layerspec));
                     }
                 }
             }
 
+            for (auto x : partial_result)
+            {
+                delete x;
+            }
+
             trace_info.file_position.pop_front();
             trace_info.selection_algorithm.pop_front();
         }
@@ -752,7 +958,9 @@ unfold_filename_iter(TraceableFileNameList& result, TraceInfo& trace_info, const
     }
     else
     {
-        result.push_back(enblend::TraceableFileName(filename, trace_info.file_position, selector));
+        result.push_back(make_traceable_file_name_and_maybe_layer(filename, trace_info.file_position,
+                                                                  selector,
+                                                                  has_layerspec, layerspec));
     }
 }
 
@@ -790,11 +998,13 @@ unfold_filename(TraceableFileNameList& result, const std::string& filename)
 #ifdef DEBUG_FILESPEC
     for (TraceableFileNameList::const_iterator i = result.begin(); i != result.end(); ++i)
     {
-        std::cout << "+ unfold_filename: filename: \"" << i->filename() <<
-            "\", selector \"" << i->selector()->name() << "\"\n";
-        for (FilePositionTrace::const_iterator j = i->trace().begin(); j != i->trace().end(); ++j)
+        std::cout <<
+            "+ unfold_filename: filename = <" << (*i)->filename() <<
+            ">, selector = " << (*i)->selector() <<
+            " -> \"" << (*i)->selector()->name() << "\"" << std::endl;
+        for (FilePositionTrace::const_iterator j = (*i)->trace().begin(); j != (*i)->trace().end(); ++j)
         {
-            std::cout << "+ unfold_filename:     pos = (" << j->first << ", " << j->second << ")\n";
+            std::cout << "+ unfold_filename: pos = (" << j->first << ", " << j->second << ")\n";
         }
     }
 #endif
@@ -805,9 +1015,7 @@ bool
 is_known_extension_to_vigra(const std::string& filename)
 {
     const std::string known_extensions = vigra::impexListExtensions();
-    std::string extension = filename.substr(filename.rfind(".") + 1);
-
-    boost::algorithm::to_lower(extension);
+    const std::string extension(enblend::to_lower_copy(filename.substr(filename.rfind(".") + 1U)));
 
     return known_extensions.find(extension) != std::string::npos;
 }
@@ -852,8 +1060,15 @@ maybe_response_file(const std::string& filename)
                 continue;
             }
 
-            if (line[0] == RESPONSE_FILE_PREFIX_CHAR ||
-                is_known_extension_to_vigra(line))
+            if (line[0] == RESPONSE_FILE_PREFIX_CHAR)
+            {
+                score += 2U;
+            }
+
+            std::string filename;
+            std::string layerspec;
+            separate_filename_and_optional_layerspec(filename, layerspec, line);
+            if (is_known_extension_to_vigra(filename))
             {
                 score += 2U;
             }
