@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2011-2012 Mikolaj Leszczynski
+ * Copyright (C) 2011-2016 Mikolaj Leszczynski
  *
  * This file is part of Enblend.
  *
@@ -17,10 +17,10 @@
  * along with Enblend; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  */
-
 #ifndef GRAPHCUT_H
 #define GRAPHCUT_H
 
+
 #include <iostream>
 
 #ifdef _WIN32
@@ -32,10 +32,7 @@
 #include <stdlib.h>
 #include <utility>
 #include <queue>
-
-#include <boost/functional/hash.hpp>
-#include <boost/scoped_ptr.hpp>
-#include <boost/unordered_set.hpp>
+#include <unordered_set>
 
 #include <vigra/functorexpression.hxx>
 #include <vigra/inspectimage.hxx>
@@ -44,14 +41,14 @@
 #include <vigra/imageiterator.hxx>
 #include <vigra/stdconvolution.hxx>
 #include <vigra/bordertreatment.hxx>
-#include <vigra/labelimage.hxx>
+#include <vigra/seededregiongrowing.hxx>
 #include <vigra/transformimage.hxx>
 #include <vigra/contourcirculator.hxx>
 #include <vigra/mathutil.hxx>
 #include <vigra/combineimages.hxx>
 
-#include "vigra_ext/rect2d.hxx"
-#include "vigra_ext/stdcachedfileimage.hxx"
+#include "rect2d.hxx"
+#include "stride.hxx"
 
 #include "common.h"
 #include "maskcommon.h"
@@ -67,8 +64,22 @@ using namespace vigra::functor;
 #define BIT_MASK_OPEN 0x04
 #define BIT_MASK_DIVIDE 0x0F
 
+#define LABEL_NONE 0
+#define LABEL_LEFT 1
+#define LABEL_RIGHT 2
 
-namespace enblend {
+
+//#define DEBUG_GRAPHCUT
+
+
+namespace enblend
+{
+    template <class T>
+    inline void hash_combine(std::size_t & seed, const T & value)
+    {
+        std::hash<T> hasher;
+        seed ^= hasher(value) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
+    }
 
     struct pointHash
     {
@@ -76,8 +87,8 @@ namespace enblend {
         {
             std::size_t seed = 0;
 
-            boost::hash_combine(seed, p.x);
-            boost::hash_combine(seed, p.y);
+            hash_combine(seed, p.x);
+            hash_combine(seed, p.y);
 
             return seed;
         }
@@ -88,7 +99,7 @@ namespace enblend {
     {
     public:
         CheckpointPixels() {}
-        boost::unordered_set<vigra::Point2D, pointHash> top, bottom;
+        std::unordered_set<vigra::Point2D, pointHash> top, bottom;
 
         ~CheckpointPixels()
         {
@@ -122,10 +133,8 @@ namespace enblend {
         typedef typename IMAGETYPE<MaskPixelType>::traverser IteratorType;
         typedef vigra::CrackContourCirculator<IteratorType> Circulator;
         typedef vigra::triple<IteratorType, IteratorType, unsigned int> EntryPointContainer;
-        IMAGETYPE<MaskPixelType> nftTempImg(mask1_lowerright - mask1_upperleft + vigra::Diff2D(2, 2),
-                                            MaskPixelTraits::max() / 2);
-        IMAGETYPE<MaskPixelType> nft(iBB.lowerRight() - iBB.upperLeft() + vigra::Diff2D(2, 2),
-                                     MaskPixelTraits::max() / 2);
+        IMAGETYPE<MaskPixelType> nftTempImg(mask1_lowerright - mask1_upperleft + vigra::Diff2D(2, 2));
+        IMAGETYPE<MaskPixelType> nft(iBB.lowerRight() - iBB.upperLeft() + vigra::Diff2D(2, 2));
         IMAGETYPE<MaskPixelType> overlap(iBB.lowerRight() - iBB.upperLeft() + vigra::Diff2D(2, 2));
         IteratorType nftIter = nft.upperLeft();
         IteratorType previous;
@@ -136,29 +145,32 @@ namespace enblend {
         bool ready = false;
         std::vector<vigra::Point2D>* interPointList = new std::vector<vigra::Point2D>();
         std::vector<EntryPointContainer> entryPointList;
-        EntryPointContainer* max = NULL;
+        EntryPointContainer* max = nullptr;
         std::pair<IteratorType, IteratorType> entryPoint;
         vigra::Point2D intermediatePoint;
         unsigned int counter = 0;
 
+        vigra::initImage(vigra::destImageRange(nftTempImg), MaskPixelTraits::max() / 2);
+        vigra::initImage(vigra::destImageRange(nft), MaskPixelTraits::max() / 2);
+
         nearestFeatureTransform(vigra::srcIterRange(mask1_upperleft, mask1_lowerright, ma1),
                                 vigra::srcIter(mask2_upperleft, ma2),
                                 vigra::destIter(nftTempImg.upperLeft() + vigra::Diff2D(1, 1)),
                                 norm, boundary);
 
-        copyImage(vigra::srcIterRange(nftTempImg.upperLeft() + vigra::Diff2D(1, 1), nftTempImg.lowerRight() - vigra::Diff2D(1, 1)),
-                  vigra::destIter(dest_upperleft, da));
+        vigra::copyImage(vigra::srcIterRange(nftTempImg.upperLeft() + vigra::Diff2D(1, 1), nftTempImg.lowerRight() - vigra::Diff2D(1, 1)),
+                         vigra::destIter(dest_upperleft, da));
 
-        copyImage(vigra_ext::apply(iBB, vigra::srcIterRange(nftTempImg.upperLeft() + vigra::Diff2D(1, 1),
-                                                     nftTempImg.lowerRight() - vigra::Diff2D(1, 1))),
-                  vigra::destIter(nft.upperLeft() + vigra::Diff2D(1, 1)));
+        vigra::copyImage(vigra_ext::apply(iBB, vigra::srcIterRange(nftTempImg.upperLeft() + vigra::Diff2D(1, 1),
+                                                                   nftTempImg.lowerRight() - vigra::Diff2D(1, 1))),
+                         vigra::destIter(nft.upperLeft() + vigra::Diff2D(1, 1)));
 
-        combineTwoImagesMP(vigra_ext::apply(iBB, vigra::srcIterRange(mask1_upperleft, mask1_lowerright, ma1)),
-                           vigra_ext::apply(iBB, vigra::srcIter(mask2_upperleft, ma2)),
-                           vigra::destIter(overlap.upperLeft() + vigra::Diff2D(1, 1)),
-                           ifThenElse(Arg1() && Arg2(),
-                                      Param(MaskPixelTraits::max()),
-                                      Param(MaskPixelTraits::zero())));
+        vigra::omp::combineTwoImages(vigra_ext::apply(iBB, vigra::srcIterRange(mask1_upperleft, mask1_lowerright, ma1)),
+                                     vigra_ext::apply(iBB, vigra::srcIter(mask2_upperleft, ma2)),
+                                     vigra::destIter(overlap.upperLeft() + vigra::Diff2D(1, 1)),
+                                     ifThenElse(Arg1() && Arg2(),
+                                                Param(MaskPixelTraits::max()),
+                                                Param(MaskPixelTraits::zero())));
 
 #ifdef DEBUG_GRAPHCUT
         exportImage(srcImageRange(nftTempImg), ImageExportInfo("./debug/nft-orig.tif").setPixelType("UINT8"));
@@ -195,7 +207,7 @@ namespace enblend {
         for (typename std::vector<EntryPointContainer>::iterator i = entryPointList.begin();
              i != entryPointList.end();
              ++i) {
-            if (max == NULL) {
+            if (max == nullptr) {
                 max = &(*i);
                 continue;
             }
@@ -204,7 +216,7 @@ namespace enblend {
             }
         }
 
-        if (max != NULL) {
+        if (max != nullptr) {
             delete circ;
             delete end;
             vigra::Diff2D dir = max->second - max->first;
@@ -328,29 +340,32 @@ namespace enblend {
     };
 
 
+    template<class MaskPixelType>
     struct OutputLabelingFunctor
     {
     public:
-        OutputLabelingFunctor(boost::unordered_set<vigra::Point2D, pointHash>* a_,
-                              boost::unordered_set<vigra::Point2D, pointHash>* b_,
+        OutputLabelingFunctor(std::unordered_set<vigra::Point2D, pointHash>* a_,
+                              std::unordered_set<vigra::Point2D, pointHash>* b_,
                               vigra::Point2D offset_) :
             left(a_), right(b_), offset(offset_) {}
 
-        bool operator()(vigra::Diff2D a2, vigra::Diff2D b2)
+        MaskPixelType operator()(const vigra::Diff2D& a_point) const
         {
-            vigra::Point2D a(a2);
-            vigra::Point2D b(b2);
-            //add border to detect seams close to border
-            a -= vigra::Point2D(1,1);
-            b -= vigra::Point2D(1,1);
-            //a-= offset; b-= offset;
-            return !((left->find(a) != left->end() && right->find(b) != right->end()) ||
-                     (right->find(a) != right->end() && left->find(b) != left->end()));
+            vigra::Point2D point(a_point);
+            // add border to detect seams close to border
+            point -= vigra::Point2D(1,1);
+
+            if(left->find(point) != left->end())
+                return LABEL_LEFT;
+            else if (right->find(point) != right->end())
+                return LABEL_RIGHT;
+            else
+                return LABEL_NONE;
         }
 
     protected:
-        boost::unordered_set<vigra::Point2D, pointHash>* left;
-        boost::unordered_set<vigra::Point2D, pointHash>* right;
+        std::unordered_set<vigra::Point2D, pointHash>* left;
+        std::unordered_set<vigra::Point2D, pointHash>* right;
         vigra::Point2D offset;
     };
 
@@ -358,8 +373,8 @@ namespace enblend {
     struct CutPixelsFunctor
     {
     public:
-        CutPixelsFunctor(boost::unordered_set<vigra::Point2D, pointHash>* a_,
-                              boost::unordered_set<vigra::Point2D, pointHash>* b_) :
+        CutPixelsFunctor(std::unordered_set<vigra::Point2D, pointHash>* a_,
+                         std::unordered_set<vigra::Point2D, pointHash>* b_) :
             left(a_), right(b_){}
 
         MaskPixelType operator()(const vigra::Diff2D& pos2, const MaskPixelType& a2) const
@@ -375,29 +390,29 @@ namespace enblend {
         }
 
     protected:
-        boost::unordered_set<vigra::Point2D, pointHash>* left;
-        boost::unordered_set<vigra::Point2D, pointHash>* right;
+        std::unordered_set<vigra::Point2D, pointHash>* left;
+        std::unordered_set<vigra::Point2D, pointHash>* right;
     };
 
     template<class MaskPixelType>
     struct CountFunctor
     {
     public:
-        CountFunctor(int* c, int* c2) : color(c), count(c2) {}
+        CountFunctor(long* c, long* c2) : color(c), count(c2) {}
 
         MaskPixelType operator()(const MaskPixelType& arg1, const MaskPixelType& arg2) const
         {
-            if (arg1 > 0 && arg1 == arg2) {
+            if (arg1 == LABEL_LEFT && arg2 == 255) {
                 (*color)++;
             }
-            if (arg1 == 255) {
+            if (arg1 == LABEL_LEFT) {
                 (*count)++;
             }
             return arg1;
         }
 
-        int* color;
-        int* count;
+        long* color;
+        long* count;
     };
 
 
@@ -479,8 +494,8 @@ namespace enblend {
 
     void
     getNeighbourList(CheckpointPixels* srcDestPoints,
-                     boost::unordered_set<vigra::Point2D, pointHash>::iterator* auxList1,
-                     boost::unordered_set<vigra::Point2D, pointHash>::iterator* auxList2)
+                     std::unordered_set<vigra::Point2D, pointHash>::iterator* auxList1,
+                     std::unordered_set<vigra::Point2D, pointHash>::iterator* auxList2)
     {
         *auxList1 = srcDestPoints->top.begin();
         *auxList2 = srcDestPoints->top.end();
@@ -558,7 +573,7 @@ namespace enblend {
     std::vector<vigra::Point2D>*
     A_star(vigra::Point2D srcpt, vigra::Point2D destpt, ImageType* img,
            GradientImageType* gradientX, GradientImageType* gradientY,
-           vigra::Diff2D bounds, CheckpointPixels* srcDestPoints, boost::unordered_set<vigra::Point2D, pointHash>* visited)
+           vigra::Diff2D bounds, CheckpointPixels* srcDestPoints, std::unordered_set<vigra::Point2D, pointHash>* visited)
     {
         MaskPixelType zeroVal = vigra::NumericTraits<MaskPixelType>::zero();
         typedef std::priority_queue<vigra::Point2D, std::vector<vigra::Point2D>, CostComparer<ImageType> > Queue;
@@ -576,8 +591,8 @@ namespace enblend {
         vigra::Point2D current;
         vigra::Point2D neighbour;
         vigra::Point2D destNeighbour;
-        boost::unordered_set<vigra::Point2D, pointHash>::iterator auxListBegin;
-        boost::unordered_set<vigra::Point2D, pointHash>::iterator auxListEnd;
+        std::unordered_set<vigra::Point2D, pointHash>::iterator auxListBegin;
+        std::unordered_set<vigra::Point2D, pointHash>::iterator auxListEnd;
         openset->push(srcpt);
 
         while (!openset->empty()) {
@@ -668,7 +683,7 @@ namespace enblend {
                     }
                 }
             } else {
-                for (boost::unordered_set<vigra::Point2D, pointHash>::iterator x = auxListBegin; x != auxListEnd; ++x) {
+                for (std::unordered_set<vigra::Point2D, pointHash>::iterator x = auxListBegin; x != auxListEnd; ++x) {
                     score = 0;
                     scoreIsBetter = false;
                     pushToList = false;
@@ -710,8 +725,8 @@ namespace enblend {
 
 
     void dividePath(std::vector<vigra::Point2D>* cut,
-                    boost::unordered_set<vigra::Point2D, pointHash>* left,
-                    boost::unordered_set<vigra::Point2D, pointHash>* right,
+                    std::unordered_set<vigra::Point2D, pointHash>* left,
+                    std::unordered_set<vigra::Point2D, pointHash>* right,
                     const vigra::Rect2D& iBB)
     {
         vigra::Point2D previous;
@@ -728,7 +743,10 @@ namespace enblend {
              ++currentDual) {
 
             current = convertFromDual(*currentDual);
-            next = convertFromDual(*nextDual);
+            
+            if (nextDual != cut->end()) {
+                next = convertFromDual(*nextDual);
+            }
 
             if (currentDual == cut->begin()) {
                 ++nextDual;
@@ -1196,82 +1214,85 @@ namespace enblend {
                       std::vector<vigra::Point2D>& totalDualPath, const vigra::Rect2D& iBB)
     {
         const vigra::Diff2D size(iBB.lowerRight().x - iBB.upperLeft().x,
-                          iBB.lowerRight().y - iBB.upperLeft().y);
+                                 iBB.lowerRight().y - iBB.upperLeft().y);
 
-        IMAGETYPE<MaskPixelType> finalmask(size + vigra::Diff2D(2, 2));
+        IMAGETYPE<MaskPixelType> finalmask(size);
         IMAGETYPE<MaskPixelType> tempImg(size);
 
         typedef vigra::UInt8 BasePixelType;
         typedef vigra::NumericTraits<BasePixelType> BasePixelTraits;
         typedef vigra::NumericTraits<MaskPixelType> MaskPixelTraits;
 
-        boost::unordered_set<vigra::Point2D, pointHash> pixelsLeftOfCut;
-        boost::unordered_set<vigra::Point2D, pointHash> pixelsRightOfCut;
+        std::unordered_set<vigra::Point2D, pointHash> pixelsLeftOfCut;
+        std::unordered_set<vigra::Point2D, pointHash> pixelsRightOfCut;
 
         dividePath(&totalDualPath, &pixelsLeftOfCut, &pixelsRightOfCut, iBB);
 
 #ifdef DEBUG_GRAPHCUT
-        combineTwoImagesMP(srcIterRange(Diff2D(), size),
-                           vigra::srcIter(finalmask.upperLeft() + Diff2D(1, 1)),
-                           vigra::destIter(tempImg.upperLeft()),
-                           CutPixelsFunctor<MaskPixelType>(&pixelsLeftOfCut, &pixelsRightOfCut));
-        exportImage(srcImageRange(tempImg), ImageExportInfo("./debug/cut_pixels.tif").setPixelType("UINT8"));
+        vigra::omp::combineTwoImages(srcIterRange(Diff2D(), size),
+                                     vigra::srcIter(finalmask.upperLeft()),
+                                     vigra::destIter(tempImg.upperLeft()),
+                                     CutPixelsFunctor<MaskPixelType>(&pixelsLeftOfCut, &pixelsRightOfCut));
+        exportImage(srcImageRange(tempImg), ImageExportInfo("./debug/process_cut_1_seam_pixels.tif").setPixelType("UINT8"));
 #endif
 
+        vigra::transformImage(srcIterRange(vigra::Diff2D(), size), vigra::destIter(tempImg.upperLeft()),
+                                OutputLabelingFunctor<MaskPixelType>(&pixelsLeftOfCut, &pixelsRightOfCut, iBB.upperLeft()));
+
         // labels areas that belong to left/right images
         // adds a 1-pixel border to catch any area that is cut off by the seam
-        vigra::labelImage(vigra::srcIterRange(vigra::Diff2D(), vigra::Diff2D() + size + vigra::Diff2D(2, 2)), destImage(finalmask),
-                          false, OutputLabelingFunctor(&pixelsLeftOfCut, &pixelsRightOfCut, iBB.upperLeft()));
+        vigra::ArrayOfRegionStatistics<vigra::SeedRgDirectValueFunctor<float> > stats(3);
+
+        vigra::seededRegionGrowing(srcImageRange(tempImg), srcImage(tempImg), destImage(finalmask), stats);
+
+        const vigra::triple<typename IMAGETYPE<MaskPixelType>::Iterator, typename IMAGETYPE<MaskPixelType>::Iterator,
+                            typename IMAGETYPE<MaskPixelType>::Accessor> finalmaskSrcRange =
+            vigra::srcIterRange(finalmask.upperLeft() + vigra::Diff2D(1, 1), finalmask.lowerRight() - vigra::Diff2D(1, 1));
+        const vigra::pair<typename IMAGETYPE<MaskPixelType>::Iterator, typename IMAGETYPE<MaskPixelType>::Accessor> finalmaskDest =
+            vigra::destIter(finalmask.upperLeft() + vigra::Diff2D(1, 1));
 
 #ifdef DEBUG_GRAPHCUT
-        exportImage(srcImageRange(finalmask), ImageExportInfo("./debug/labels.tif").setPixelType("UINT8"));
+        exportImage(srcImageRange(tempImg), ImageExportInfo("./debug/process_cut_2_seam_labels.tif").setPixelType("UINT8"));
+        exportImage(finalmaskSrcRange, ImageExportInfo("./debug/process_cut_3_segments.tif").setPixelType("UINT8"));
 #endif
 
-        int colorSum = 0;
-        int count = 0;
-        CountFunctor<MaskPixelType> c(&colorSum, &count);
+        long colorSum = 0;
+        long count = 0;
+        CountFunctor<MaskPixelType> countFunctor(&colorSum, &count);
 
-        transformImage(vigra::srcIterRange(finalmask.upperLeft() + vigra::Diff2D(1, 1),
-                                    finalmask.lowerRight() - vigra::Diff2D(1, 1)),
-                       vigra::destIter(finalmask.upperLeft() + vigra::Diff2D(1, 1)),
-                       ifThenElse(Arg1() == Param(1), Param(BasePixelTraits::max()),
-                                  Param(BasePixelTraits::zero())));
+        inspectTwoImages(finalmaskSrcRange,
+                         vigra::srcIter(dest_upperleft + iBB.upperLeft(), da), countFunctor);
 
-        inspectTwoImages(vigra::srcIterRange(finalmask.upperLeft() + vigra::Diff2D(1, 1),
-                                      finalmask.lowerRight() - vigra::Diff2D(1, 1)),
-                         vigra::srcIter(dest_upperleft + iBB.upperLeft(), da), c);
-#ifdef DEBUG_GRAPHCUT
-        exportImage(srcImageRange(finalmask), ImageExportInfo("./debug/labels1.tif").setPixelType("UINT8"));
-#endif
-
-        // if colorSum < half the pixels labeled as "1" in finalmask should be black
+        // if colorSum < half the pixels labeled as LABEL_LEFT in finalmask should be white
         if (colorSum < count / 2) {
-            transformImage(vigra::srcIterRange(finalmask.upperLeft() + vigra::Diff2D(1, 1),
-                                        finalmask.lowerRight() - vigra::Diff2D(1, 1)),
-                           vigra::destIter(finalmask.upperLeft() + vigra::Diff2D(1, 1)),
-                           ifThenElse(Arg1() == Param(0), Param(BasePixelTraits::max()),
-                                      Param(BasePixelTraits::zero())));
-        } else {
-            transformImage(vigra::srcIterRange(finalmask.upperLeft() + vigra::Diff2D(1, 1),
-                                        finalmask.lowerRight() - vigra::Diff2D(1, 1)),
-                           vigra::destIter(finalmask.upperLeft() + vigra::Diff2D(1, 1)),
-                           ifThenElse(Arg1() == Param(0), Param(BasePixelTraits::zero()),
+            transformImage(finalmaskSrcRange, finalmaskDest,
+                           ifThenElse(Arg1() == Param(LABEL_LEFT),
+                                      Param(BasePixelTraits::zero()),
                                       Param(BasePixelTraits::max())));
+        } else {
+            transformImage(finalmaskSrcRange, finalmaskDest,
+                           ifThenElse(Arg1() == Param(LABEL_LEFT),
+                                      Param(BasePixelTraits::max()),
+                                      Param(BasePixelTraits::zero())));
         }
 
 #ifdef DEBUG_GRAPHCUT
-        exportImage(srcImageRange(finalmask), ImageExportInfo("./debug/labels2.tif").setPixelType("UINT8"));
+        exportImage(finalmaskSrcRange, ImageExportInfo("./debug/process_cut_4_final_precopy.tif").setPixelType("UINT8"));
 #endif
 
-        combineTwoImagesMP(vigra_ext::apply(iBB, vigra::srcIterRange(mask1_upperleft, mask1_lowerright, ma1)),
-                           vigra_ext::apply(iBB, vigra::srcIter(mask2_upperleft, ma2)), destImage(tempImg),
-                           ifThenElse(Arg1() && Arg2(),
-                                      Param(MaskPixelTraits::max()),
-                                      Param(MaskPixelTraits::zero())));
+        vigra::omp::combineTwoImages(vigra_ext::apply(iBB, vigra::srcIterRange(mask1_upperleft, mask1_lowerright, ma1)),
+                                     vigra_ext::apply(iBB, vigra::srcIter(mask2_upperleft, ma2)), destImage(tempImg),
+                                     ifThenElse(Arg1() && Arg2(),
+                                                Param(MaskPixelTraits::max()),
+                                                Param(MaskPixelTraits::zero())));
+
+        vigra::copyImageIf(finalmaskSrcRange, srcImage(tempImg),
+                           vigra::destIter(dest_upperleft + iBB.upperLeft(), da));
+
+#ifdef DEBUG_GRAPHCUT
+        exportImage(srcIterRange(dest_upperleft + iBB.upperLeft(), dest_upperleft + iBB.upperLeft() + size), ImageExportInfo("./debug/process_cut_5_final_postcopy.tif").setPixelType("UINT8"));
+#endif
 
-        copyImageIf(vigra::srcIterRange(finalmask.upperLeft() + vigra::Diff2D(1, 1),
-                                 finalmask.lowerRight() - vigra::Diff2D(1, 1)), srcImage(tempImg),
-                    vigra::destIter(dest_upperleft + iBB.upperLeft(), da));
     }
 
 
@@ -1301,7 +1322,6 @@ namespace enblend {
         typedef vigra::UInt8 BasePixelType;
         typedef float GradientPixelType;
         typedef vigra::NumericTraits<BasePixelType> BasePixelTraits;
-        typedef vigra::NumericTraits<GradientPixelType> GradientPixelTraits;
         typedef typename BasePixelTraits::Promote BasePromotePixelType;
         typedef vigra::NumericTraits<BasePromotePixelType> BasePromotePixelTraits;
         typedef typename BasePromotePixelTraits::Promote GraphPixelType;
@@ -1320,10 +1340,10 @@ namespace enblend {
         IMAGETYPE<GradientPixelType> gradientY(size);
         IMAGETYPE<GraphPixelType> graphImg(size + size + vigra::Diff2D(1, 1));
 
-        std::vector<vigra::Point2D>* dualPath = NULL;
+        std::vector<vigra::Point2D>* dualPath = nullptr;
         std::vector<vigra::Point2D> totalDualPath;
         vigra::Point2D intermediatePoint;
-        boost::scoped_ptr<CheckpointPixels> srcDestPoints(new CheckpointPixels());
+        CheckpointPixels srcDestPoints;
 
         const vigra::Diff2D graphsize(graphImg.lowerRight().x - graphImg.upperLeft().x,
                                graphImg.lowerRight().y - graphImg.upperLeft().y);
@@ -1346,10 +1366,10 @@ namespace enblend {
         transformImage(src2_upperleft, src2_upperleft + size, sa2,
                        intermediateImg.upperLeft(), intermediateImg.accessor(),
                        MapFunctor<SrcPixelType, BasePixelType>());
-        combineTwoImagesMP(srcImageRange(gradientPreConvolve),
-                           srcImage(intermediateImg),
-                           destImage(gradientPreConvolve),
-                           Arg1() + Arg2());
+        vigra::omp::combineTwoImages(srcImageRange(gradientPreConvolve),
+                                     srcImage(intermediateImg),
+                                     destImage(gradientPreConvolve),
+                                     Arg1() + Arg2());
 
         // computing image gradient for a better cost function
         vigra::separableConvolveX(srcImageRange(gradientPreConvolve),
@@ -1363,29 +1383,29 @@ namespace enblend {
         switch (PixelDifferenceFunctor)
         {
         case HueLuminanceMaxDifference:
-            combineTwoImagesMP(src1_upperleft, src1_lowerright, sa1,
-                               src2_upperleft, sa2,
-                               intermediateImg.upperLeft(), intermediateImg.accessor(),
-                               MaxHueLuminanceDifferenceFunctor<SrcPixelType, BasePixelType>
-                               (LuminanceDifferenceWeight, ChrominanceDifferenceWeight));
+            vigra::omp::combineTwoImages(src1_upperleft, src1_lowerright, sa1,
+                                         src2_upperleft, sa2,
+                                         intermediateImg.upperLeft(), intermediateImg.accessor(),
+                                         MaxHueLuminanceDifferenceFunctor<SrcPixelType, BasePixelType>
+                                         (LuminanceDifferenceWeight, ChrominanceDifferenceWeight));
             break;
         case DeltaEDifference:
-            combineTwoImagesMP(src1_upperleft, src1_lowerright, sa1,
-                               src2_upperleft, sa2,
-                               intermediateImg.upperLeft(), intermediateImg.accessor(),
-                               DeltaEPixelDifferenceFunctor<SrcPixelType, BasePixelType>
-                               (LuminanceDifferenceWeight, ChrominanceDifferenceWeight));
+            vigra::omp::combineTwoImages(src1_upperleft, src1_lowerright, sa1,
+                                         src2_upperleft, sa2,
+                                         intermediateImg.upperLeft(), intermediateImg.accessor(),
+                                         DeltaEPixelDifferenceFunctor<SrcPixelType, BasePixelType>
+                                         (LuminanceDifferenceWeight, ChrominanceDifferenceWeight));
             break;
         default:
-            throw never_reached("switch control expression \"PixelDifferenceFunctor\" out of range");
+            NEVER_REACHED("switch control expression \"PixelDifferenceFunctor\" out of range");
         }
 
         // masking overlap region borders
-        combineThreeImagesMP(vigra_ext::apply(iBB, vigra::srcIterRange(mask1_upperleft, mask1_lowerright, ma1)),
-                             vigra_ext::apply(iBB, vigra::srcIter(mask2_upperleft, ma2)),
-                             vigra::srcIter(intermediateImg.upperLeft(), intermediateImg.accessor()),
-                             vigra::destIter(intermediateImg.upperLeft(), intermediateImg.accessor()),
-                             ifThenElse(!(Arg1() & Arg2()), Param(BasePixelTraits::max()), Arg3()));
+        vigra::omp::combineThreeImages(vigra_ext::apply(iBB, vigra::srcIterRange(mask1_upperleft, mask1_lowerright, ma1)),
+                                       vigra_ext::apply(iBB, vigra::srcIter(mask2_upperleft, ma2)),
+                                       vigra::srcIter(intermediateImg.upperLeft(), intermediateImg.accessor()),
+                                       vigra::destIter(intermediateImg.upperLeft(), intermediateImg.accessor()),
+                                       ifThenElse(!(Arg1() & Arg2()), Param(BasePixelTraits::max()), Arg3()));
 
         // look for possible start and end points
         std::vector<vigra::Point2D>* intermediatePointList =
@@ -1401,12 +1421,12 @@ namespace enblend {
         }
 
         // copying to a grid
-        copyImage(srcImageRange(intermediateImg),
-                  vigra_ext::stride(2, 2, vigra_ext::apply(gBB, destImage(intermediateGraphImg))));
+        vigra::copyImage(srcImageRange(intermediateImg),
+                         vigra_ext::stride(2, 2, vigra_ext::apply(gBB, destImage(intermediateGraphImg))));
 
         // calculating differences between pixels that are adjacent in the original image
         convolveImage(srcImageRange(intermediateGraphImg), destImage(graphImg), vigra::kernel2d(edgeWeightKernel));
-        copyImage(srcImageRange(graphImg), destImage(intermediateGraphImg));
+        vigra::copyImage(srcImageRange(graphImg), destImage(intermediateGraphImg));
 
 #ifdef DEBUG_GRAPHCUT
         exportImage(srcImageRange(intermediateImg), ImageExportInfo("./debug/diff.tif").setPixelType("UINT8"));
@@ -1421,7 +1441,7 @@ namespace enblend {
 #endif
 
         // a set of points to keep visited points for subsequent graph-cut runs
-        boost::unordered_set<vigra::Point2D, pointHash> visited;
+        std::unordered_set<vigra::Point2D, pointHash> visited;
 
         // find optimal cuts in dual graph
         for (std::vector<vigra::Point2D>::iterator i = intermediatePointList->begin();
@@ -1432,9 +1452,9 @@ namespace enblend {
                 continue;
             }
 
-            srcDestPoints->clear();
-            srcDestPoints->top.insert(intermediatePoint);
-            srcDestPoints->bottom.insert(*i);
+            srcDestPoints.clear();
+            srcDestPoints.top.insert(intermediatePoint);
+            srcDestPoints.bottom.insert(*i);
 
 #ifdef DEBUG_GRAPHCUT
             std::cout << "Running graph-cut: " << intermediatePoint << ":" << *i << std::endl;
@@ -1442,7 +1462,7 @@ namespace enblend {
 
             dualPath = A_star<IMAGETYPE<GraphPixelType>, IMAGETYPE<GradientPixelType>, BasePixelType>
                 (vigra::Point2D(-10, -10), vigra::Point2D(-20, -20), &intermediateGraphImg, &gradientX,
-                 &gradientY, graphsize - vigra::Diff2D(1, 1), srcDestPoints.get(), &visited);
+                 &gradientY, graphsize - vigra::Diff2D(1, 1), &srcDestPoints, &visited);
 
             visited.insert(dualPath->begin(), dualPath->end());
 
@@ -1452,7 +1472,7 @@ namespace enblend {
                 }
             }
 
-            copyImage(srcImageRange(graphImg), destImage(intermediateGraphImg));
+            vigra::copyImage(srcImageRange(graphImg), destImage(intermediateGraphImg));
             intermediatePoint = *i;
         }
 
@@ -1463,10 +1483,12 @@ namespace enblend {
         delete intermediatePointList;
         delete dualPath;
     }
-
 } /* namespace enblend */
+
+
 #endif  /* GRAPHCUT_H */
 
+
 // Local Variables:
 // mode: c++
 // End:
