@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2012 Christoph L. Spiel
+ * Copyright (C) 2008-2016 Christoph L. Spiel
  *
  * This file is part of Enblend.
  *
@@ -63,7 +63,9 @@ public:
         return f(i, d, srcIsScalar());
     }
 
-    static const std::string defaultGrayscaleAccessorName() {return "average";}
+    static const std::string defaultGrayscaleAccessorName() {
+        return "average";       //< default-grayscale-accessor average
+    }
 
 private:
     typedef enum AccessorKind {
@@ -131,7 +133,7 @@ private:
                     std::cerr << command
                          << ": \"" CHANNEL_MIXER "\" is a grayscale projector requiring\n"
                          << command
-                         << ":      arguments like e.g. \"channel-mixer:0.30:0.59:0.11\""
+                         << ": arguments like e.g. \"channel-mixer:0.30:0.59:0.11\""
                          << std::endl;
                     exit(1);
                 }
