@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2012 Christoph L. Spiel
+ * Copyright (C) 2009-2016 Christoph L. Spiel
  *
  * This file is part of Enblend.
  *
@@ -37,7 +37,98 @@
 #include <vigra/inspectimage.hxx>
 #include <vigra/numerictraits.hxx>
 
-#include "vigra_ext/stdcachedfileimage.hxx"
+#include "timer.h"
+#include "opencl_vigra.h"
+
+
+#ifdef OPENCL
+namespace GPU
+{
+    extern std::unique_ptr<vigra::ocl::DistanceTransformFH> DistanceTransform;
+}
+#endif
+
+
+namespace vigra
+{
+    namespace ocl
+    {
+        template <class SrcImageIterator, class SrcAccessor,
+                  class DestImageIterator, class DestAccessor,
+                  class ValueType>
+        inline void
+        distanceTransform(SrcImageIterator src_upperleft, SrcImageIterator src_lowerright, SrcAccessor src_acc,
+                          DestImageIterator dest_upperleft, DestAccessor dest_acc,
+                          ValueType background, int norm)
+        {
+            timer::WallClock wall_clock;
+
+            wall_clock.start();
+
+#ifdef OPENCL
+            const bool enable_kernel = parameter::as_boolean("gpu-kernel-dt", true);
+
+            if (GPUContext && GPU::DistanceTransform && enable_kernel)
+            {
+#ifdef DEBUG
+                std::cerr <<
+                    command << ": info: choose OpenCL accelaration for Distance Transform" << std::endl;
+#endif // DEBUG
+
+                GPU::DistanceTransform->run(src_upperleft, src_lowerright, src_acc,
+                                            dest_upperleft, dest_acc,
+                                            background, norm);
+            }
+            else
+            {
+                if (UseGPU && enable_kernel)
+                {
+                    std::cerr <<
+                        command << ": warning: missing GPUContext or OpenCL DistanceTransform\n" <<
+                        command << ": warning: falling back to CPU path" << std::endl;
+                }
+
+                vigra::omp::distanceTransform(src_upperleft, src_lowerright, src_acc,
+                                              dest_upperleft, dest_acc,
+                                              background, norm);
+            }
+#else
+            vigra::omp::distanceTransform(src_upperleft, src_lowerright, src_acc,
+                                          dest_upperleft, dest_acc,
+                                          background, norm);
+#endif // OPENCL
+
+            wall_clock.stop();
+            if (parameter::as_boolean("time-distance-transform", false))
+            {
+                const std::ios::fmtflags flags(std::cerr.flags());
+                vigra::Size2D size(src_lowerright - src_upperleft);
+                std::cerr <<
+                    "\n" <<
+                    command << ": timing: wall-clock runtime of `Distance Transform': " <<
+                    std::setprecision(3) << 1000.0 * wall_clock.value() << " ms\n" <<
+                    command << ": timing: speed according to wall-clock: " <<
+                    size.area() / (1048576.0 * wall_clock.value()) << " MPixel/s\n" <<
+                    std::endl;
+                std::cerr.flags(flags);
+            }
+        }
+
+
+        template <class SrcImageIterator, class SrcAccessor,
+                  class DestImageIterator, class DestAccessor,
+                  class ValueType>
+        inline static void
+        distanceTransform(vigra::triple<SrcImageIterator, SrcImageIterator, SrcAccessor> src,
+                          vigra::pair<DestImageIterator, DestAccessor> dest,
+                          ValueType background, int norm)
+        {
+            vigra::ocl::distanceTransform(src.first, src.second, src.third,
+                                          dest.first, dest.second,
+                                          background, norm);
+        }
+    } // namespace ocl
+} // namespace vigra
 
 
 namespace enblend {
@@ -56,37 +147,37 @@ quadruple_image(SrcImageIterator src_upperleft,
     switch (boundary)
     {
     case OpenBoundaries:
-        copyImage(src_upperleft, src_lowerright, sa,
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
                   dest_upperleft, da);
         break;
 
     case HorizontalStrip:
-        copyImage(src_upperleft, src_lowerright, sa,
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
                   dest_upperleft, da); // 11
-        copyImage(src_upperleft, src_lowerright, sa,
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
                   dest_upperleft + size_x, da); // 12
         break;
 
     case VerticalStrip:
-        copyImage(src_upperleft, src_lowerright, sa,
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
                   dest_upperleft, da); // 11
-        copyImage(src_upperleft, src_lowerright, sa,
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
                   dest_upperleft + size_y, da); // 21
         break;
 
     case DoubleStrip:
-        copyImage(src_upperleft, src_lowerright, sa,
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
                   dest_upperleft, da); // 11
-        copyImage(src_upperleft, src_lowerright, sa,
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
                   dest_upperleft + size_x, da); // 12
-        copyImage(src_upperleft, src_lowerright, sa,
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
                   dest_upperleft + size_y, da); // 21
-        copyImage(src_upperleft, src_lowerright, sa,
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
                   dest_upperleft + size_x + size_y, da); // 22
         break;
 
     default:
-        throw never_reached("switch control expression \"boundary\" out of range");
+        NEVER_REACHED("switch control expression \"boundary\" out of range");
     }
 }
 
@@ -127,61 +218,61 @@ quater_image(SrcImageIterator src_upperleft,
     switch (boundary)
     {
     case OpenBoundaries:
-        copyImage(src_upperleft, src_lowerright, sa,
-                  dest_upperleft, da);
+        vigra::copyImage(src_upperleft, src_lowerright, sa,
+                         dest_upperleft, da);
         break;
 
     case HorizontalStrip:
-        copyImage(src_upperleft + size_x2,
-                  src_upperleft + size_x2 + size_x4 + size_y,
-                  sa,
-                  dest_upperleft,
-                  da); // 11
-        copyImage(src_upperleft + size_x4,
-                  src_upperleft + size_x2 + size_y,
-                  sa,
-                  dest_upperleft + size_x4,
-                  da); // 12
+        vigra::copyImage(src_upperleft + size_x2,
+                         src_upperleft + size_x2 + size_x4 + size_y,
+                         sa,
+                         dest_upperleft,
+                         da); // 11
+        vigra::copyImage(src_upperleft + size_x4,
+                         src_upperleft + size_x2 + size_y,
+                         sa,
+                         dest_upperleft + size_x4,
+                         da); // 12
         break;
 
     case VerticalStrip:
-        copyImage(src_upperleft + size_y2,
-                  src_upperleft + size_y2 + size_y4 + size_x,
-                  sa,
-                  dest_upperleft,
-                  da); // 21
-        copyImage(src_upperleft + size_y4,
-                  src_upperleft + size_y2 + size_x,
-                  sa,
-                  dest_upperleft + size_y4,
-                  da); // 22
+        vigra::copyImage(src_upperleft + size_y2,
+                         src_upperleft + size_y2 + size_y4 + size_x,
+                         sa,
+                         dest_upperleft,
+                         da); // 21
+        vigra::copyImage(src_upperleft + size_y4,
+                         src_upperleft + size_y2 + size_x,
+                         sa,
+                         dest_upperleft + size_y4,
+                         da); // 22
         break;
 
     case DoubleStrip:
-        copyImage(src_upperleft + size_x2 + size_y2,
-                  src_upperleft + size_x2 + size_y2 + size_x4 + size_y4,
-                  sa,
-                  dest_upperleft,
-                  da); // 11
-        copyImage(src_upperleft + size_x4 + size_y2,
-                  src_upperleft + size_x2 + size_y2 + size_y4,
-                  sa,
-                  dest_upperleft + size_x4,
-                  da); // 12
-        copyImage(src_upperleft + size_x2 + size_y4,
-                  src_upperleft + size_x2 + size_x4 + size_y2,
-                  sa,
-                  dest_upperleft + size_y4,
-                  da); // 21
-        copyImage(src_upperleft + size_x4 + size_y4,
-                  src_upperleft + size_x2 + size_y2,
-                  sa,
-                  dest_upperleft + size_x4 + size_y4,
-                  da); // 22
+        vigra::copyImage(src_upperleft + size_x2 + size_y2,
+                         src_upperleft + size_x2 + size_y2 + size_x4 + size_y4,
+                         sa,
+                         dest_upperleft,
+                         da); // 11
+        vigra::copyImage(src_upperleft + size_x4 + size_y2,
+                         src_upperleft + size_x2 + size_y2 + size_y4,
+                         sa,
+                         dest_upperleft + size_x4,
+                         da); // 12
+        vigra::copyImage(src_upperleft + size_x2 + size_y4,
+                         src_upperleft + size_x2 + size_x4 + size_y2,
+                         sa,
+                         dest_upperleft + size_y4,
+                         da); // 21
+        vigra::copyImage(src_upperleft + size_x4 + size_y4,
+                         src_upperleft + size_x2 + size_y2,
+                         sa,
+                         dest_upperleft + size_x4 + size_y4,
+                         da); // 22
         break;
 
     default:
-        throw never_reached("switch control expression \"boundary\" out of range");
+        NEVER_REACHED("switch control expression \"boundary\" out of range");
     }
 }
 
@@ -238,17 +329,14 @@ periodicDistanceTransform(SrcImageIterator src_upperleft, SrcImageIterator src_l
         break;
 
     default:
-        throw never_reached("switch control expression \"boundary\" out of range");
+        NEVER_REACHED("switch control expression \"boundary\" out of range");
     }
 
     vigra::BasicImage<SrcValueType> periodic(size_x, size_y);
     vigra::BasicImage<DestValueType> distance(periodic.size());
 
-    quadruple_image(src_upperleft, src_lowerright, sa,
-                    periodic.upperLeft(), periodic.accessor(),
-                    boundary);
-    distanceTransformMP(srcImageRange(periodic), destImage(distance),
-                        background, norm);
+    quadruple_image(src_upperleft, src_lowerright, sa, periodic.upperLeft(), periodic.accessor(), boundary);
+    vigra::ocl::distanceTransform(srcImageRange(periodic), destImage(distance), background, norm);
     quater_image(srcImageRange(distance), destIter(dest_upperleft, da), boundary);
 }
 
@@ -343,17 +431,15 @@ nearestFeatureTransform(SrcImageIterator src1_upperleft, SrcImageIterator src1_l
 {
     typedef typename SrcAccessor::value_type SrcPixelType;
     typedef vigra::NumericTraits<SrcPixelType> SrcPixelTraits;
-    typedef typename SrcPixelTraits::Promote SrcPromoteType;
 
     typedef typename DestAccessor::value_type DestPixelType;
     typedef vigra::NumericTraits<DestPixelType> DestPixelTraits;
 
     const SrcPixelType background = SrcPixelTraits::zero();
-    const vigra::Diff2D size(src1_lowerright.x - src1_upperleft.x,
-                             src1_lowerright.y - src1_upperleft.y);
+    const vigra::Size2D size(src1_lowerright.x - src1_upperleft.x, src1_lowerright.y - src1_upperleft.y);
 
-    IMAGETYPE<SrcPromoteType> dist12(size);
-    IMAGETYPE<SrcPromoteType> dist21(size);
+    IMAGETYPE<float> dist12(size);
+    IMAGETYPE<float> dist21(size);
     if (Verbose >= VERBOSE_NFT_MESSAGES)
     {
         std::cerr << command << ": info: creating ";
@@ -371,85 +457,64 @@ nearestFeatureTransform(SrcImageIterator src1_upperleft, SrcImageIterator src1_l
         // consider the overlap of the masks to be complete, i.e. the
         // image pair as useless.
         //
-        // The current value is 2x the circumference of the overlap
-        // rectangle.
-        2U * 2U * (static_cast<unsigned>(size.x) + static_cast<unsigned>(size.y));
-    unsigned tally12;
-    unsigned tally21;
+        // The current parameter default is two times the
+        // circumference of the overlap rectangle.
+        parameter::as_unsigned("overlap-check-threshold", 2U) * //< overlap-check-threshold 2
+        2U * (static_cast<unsigned>(size.height()) + static_cast<unsigned>(size.width()));
+    IMAGETYPE<SrcPixelType> diff12(size);
+    vigra::omp::combineTwoImages(src1_upperleft, src1_lowerright, sa1,
+                                 src2_upperleft, sa2,
+                                 diff12.upperLeft(), diff12.accessor(),
+                                 saturating_subtract<SrcPixelType>());
 
-#ifdef OPENMP
-#pragma omp parallel sections
-#endif
+    const unsigned tally12 = quick_tally(diff12.begin(), diff12.end(), diff12.accessor(), overlap_threshold);
+
+    switch (boundary)
     {
-#ifdef OPENMP
-#pragma omp section
-#endif
-        {
-            IMAGETYPE<SrcPixelType> diff12(size);
-            combineTwoImagesMP(src1_upperleft, src1_lowerright, sa1,
-                               src2_upperleft, sa2,
-                               diff12.upperLeft(), diff12.accessor(),
-                               saturating_subtract<SrcPixelType>());
+    case OpenBoundaries:
+        vigra::ocl::distanceTransform(srcImageRange(diff12), destImage(dist12), background, norm);
+        break;
 
-            tally12 = quick_tally(diff12.begin(), diff12.end(), diff12.accessor(),
-                                  overlap_threshold);
+    case HorizontalStrip: // FALLTHROUGH
+    case VerticalStrip:   // FALLTHROUGH
+    case DoubleStrip:
+        periodicDistanceTransform(srcImageRange(diff12), destImage(dist12),
+                                  background, norm, boundary);
+        break;
 
-            switch (boundary)
-            {
-            case OpenBoundaries:
-                distanceTransformMP(srcImageRange(diff12), destImage(dist12),
-                                    background, norm);
-                break;
+    default:
+        NEVER_REACHED("switch control expression \"boundary\" out of range");
+    }
 
-            case HorizontalStrip: // FALLTHROUGH
-            case VerticalStrip:   // FALLTHROUGH
-            case DoubleStrip:
-                periodicDistanceTransform(srcImageRange(diff12), destImage(dist12),
-                                          background, norm, boundary);
-                break;
+    if (Verbose >= VERBOSE_NFT_MESSAGES)
+    {
+        std::cerr << " 2/3";
+        std::cerr.flush();
+    }
+    IMAGETYPE<SrcPixelType> diff21(size);
+    vigra::omp::combineTwoImages(src2_upperleft, src2_upperleft + size, sa2,
+                                 src1_upperleft, sa1,
+                                 diff21.upperLeft(), diff21.accessor(),
+                                 saturating_subtract<SrcPixelType>());
 
-            default:
-                throw never_reached("switch control expression \"boundary\" out of range");
-            }
-        } // omp section
+    const unsigned tally21 = quick_tally(diff21.begin(), diff21.end(), diff21.accessor(), overlap_threshold);
 
-#ifdef OPENMP
-#pragma omp section
-#endif
-        {
-            if (Verbose >= VERBOSE_NFT_MESSAGES)
-            {
-                std::cerr << " 2/3";
-                std::cerr.flush();
-            }
-            IMAGETYPE<SrcPixelType> diff21(size);
-            combineTwoImagesMP(src2_upperleft, src2_upperleft + size, sa2,
-                               src1_upperleft, sa1,
-                               diff21.upperLeft(), diff21.accessor(),
-                               saturating_subtract<SrcPixelType>());
+    switch (boundary)
+    {
+    case OpenBoundaries:
+        vigra::ocl::distanceTransform(srcImageRange(diff21), destImage(dist21), background, norm);
+        break;
 
-            tally21 = quick_tally(diff21.begin(), diff21.end(), diff21.accessor(),
-                                  overlap_threshold);
+    case HorizontalStrip: // FALLTHROUGH
+    case VerticalStrip:   // FALLTHROUGH
+    case DoubleStrip:
+        periodicDistanceTransform(srcImageRange(diff21), destImage(dist21),
+                                  background, norm, boundary);
+        break;
 
-            switch (boundary)
-            {
-            case OpenBoundaries:
-                distanceTransformMP(srcImageRange(diff21), destImage(dist21),
-                                    background, norm);
-                break;
-
-            case HorizontalStrip: // FALLTHROUGH
-            case VerticalStrip:   // FALLTHROUGH
-            case DoubleStrip:
-                periodicDistanceTransform(srcImageRange(diff21), destImage(dist21),
-                                          background, norm, boundary);
-                break;
-
-            default:
-                throw never_reached("switch control expression \"boundary\" out of range");
-            }
-        } // omp section
-    } // omp parallel sections
+    default:
+        NEVER_REACHED("switch control expression \"boundary\" out of range");
+    }
 
     if (Verbose >= VERBOSE_NFT_MESSAGES)
     {
@@ -461,24 +526,23 @@ nearestFeatureTransform(SrcImageIterator src1_upperleft, SrcImageIterator src1_l
     if (overlap_tally < overlap_threshold)
     {
         std::cerr << "\n" <<
-            command << ": excessive overlap detected; remove one of the images\n";
-#ifdef DEBUG_NEAREST_FEATURE_TRANSFORM
-        cout <<
-            "+ nearestFeatureTransform: overlap area size = " << size << "\n" <<
-            "+ nearestFeatureTransform: threshold is " << overlap_threshold <<
-            " pixels for this pair of images\n" <<
-            "+ nearestFeatureTransform: only " << overlap_tally << " of " <<
-            size.x * size.y << " pixels do not overlap\n";
+            command << ": excessive image overlap detected; too high risk of defective seam line\n" <<
+            command << ": note: remove at least one of the images" << std::endl;
+#ifdef DEBUG
+        std::cerr <<
+            command << ": note: overlap area size is " << size << " > " << overlap_threshold << ", but only\n" <<
+            command << ": note: " << overlap_tally << " of " << size.x * size.y << " pixels do not overlap" <<
+            std::endl;
 #endif
         exit(1);
     }
 
-    combineTwoImagesMP(dist12.upperLeft(), dist12.lowerRight(), dist12.accessor(),
-                       dist21.upperLeft(), dist21.accessor(),
-                       dest_upperleft, da,
-                       ifThenElse(vigra::functor::Arg1() < vigra::functor::Arg2(),
-                                  vigra::functor::Param(DestPixelTraits::max()),
-                                  vigra::functor::Param(DestPixelTraits::zero())));
+    vigra::omp::combineTwoImages(dist12.upperLeft(), dist12.lowerRight(), dist12.accessor(),
+                                 dist21.upperLeft(), dist21.accessor(),
+                                 dest_upperleft, da,
+                                 ifThenElse(vigra::functor::Arg1() < vigra::functor::Arg2(),
+                                            vigra::functor::Param(DestPixelTraits::max()),
+                                            vigra::functor::Param(DestPixelTraits::zero())));
 
     if (Verbose >= VERBOSE_NFT_MESSAGES)
     {
